/* $Header: /cvsroot/dhcp-agent/dhcp-agent/src/dhcp-client-conf.c,v 1.29 2003/06/27 03:16:27 actmodern Exp $
 * 
 * Copyright 2002 Thamer Alharbash <tmh@whitefang.com>
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in
 * the documentation and/or other materials provided with the
 * distribution.
 * 3. The names of the authors may not be used to endorse or promote
 * products derived from this software without specific prior
 * written permission.
 * 
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 * 
 */

#define MODULE_NAME "dhcp-client-conf"

#include "dhcp-local.h"
#include "dhcp-limits.h"
#include "dhcp-libutil.h"
#include "dhcp-librawnet.h"
#include "dhcp-option.h"

#include "dhcp-conf.h"
#include "dhcp-conf-var.h"
#include "dhcp-client-conf.h"
#include "dhcp-cache-entry.h"
#include "dhcp-client-cache.h"
#include "dhcp-options-strings.h"

#include "dhcp-client-defaults.h"

/* * * * * * * * * * * * * * * * * * *
 * configuration grammar structures. *
 * * * * * * * * * * * * * * * * * * */

/*
 * We construct the tree for client-conf code to handle the
 * grammar for us.
 *
 * The one limitation we have now is that the tree has branches
 * from the top, the command name itself, and everything else is
 * linear. This is fine for what we want to do now but it may
 * need to support a more proper grammer tree for more complex
 * parsing in the future. */

/* string arrays for commands. */

/* variables the client will accept to be set. */
static const char *var_strings[] = {
    "hostname",

    /* protocol related. */

    /* timeout thresholds. */
    "dhcp-discover-timeout-threshold",
    "dhcp-request-timeout-threshold",
    "arp-timeout-threshold",
    "icmp-echo-timeout-threshold",
    "icmp-subnet-timeout-threshold",

    /* retries. */
    "dhcp-discover-retries",
    "dhcp-request-retries",
    "icmp-retries",
    "arp-retries",

    "default-interface-mtu",
    "default-subnet-mask",

    /* lease time default percent */
    "default-renew-percent",
    "default-rebind-percent",

};

static const char *var_boolean_strings[] = {
    "do-measure-router-latency",
};

/* strings for ip address or mac address with server command. */
static const char *server_strings[] = {
    "ip-address",
    "mac-address",
};

/* symbol arrays. */

/* symbols for string substitution. */
static const arg_symbol_t var_symbols[] = {
    CLIENT_VAR_HOSTNAME,
    CLIENT_VAR_DHCP_DISCOVER_TIMEOUT,
    CLIENT_VAR_DHCP_REQUEST_TIMEOUT,
    CLIENT_VAR_ARP_TIMEOUT,
    CLIENT_VAR_ICMP_ECHO_TIMEOUT,
    CLIENT_VAR_ICMP_SUBNET_TIMEOUT,
    CLIENT_VAR_DHCP_DISCOVER_RETRIES,
    CLIENT_VAR_DHCP_REQUEST_RETRIES,
    CLIENT_VAR_ICMP_RETRIES,
    CLIENT_VAR_ARP_RETRIES,
    CLIENT_VAR_INTERFACE_MTU,
    CLIENT_VAR_SUBNET_MASK,
    CLIENT_VAR_RENEW_PERCENT,
    CLIENT_VAR_REBIND_PERCENT,
};

static const arg_symbol_t server_symbols[] = {
    CLIENT_SERVER_IP,
    CLIENT_SERVER_MAC,
};

static const arg_symbol_t var_boolean_symbols[] = {
    CLIENT_VAR_DO_MEASURE_ROUTER_LATENCY_ICMP,
};

/* argument types, string array pointers, and symbol array pointers */

static const arg_type_t     var_boolean_arg_types[] = { CONF_IDENTIFIER, CONF_ASSIGNMENT, CONF_BOOLEAN };
static const char         **var_boolean_arg_strings[] = { var_boolean_strings, NULL, NULL };
static const arg_symbol_t  *var_boolean_arg_symbols[] = { var_boolean_symbols, NULL, NULL };

static const arg_type_t     var_arg_types[] = { CONF_IDENTIFIER, CONF_ASSIGNMENT, CONF_STRING };
static const char         **var_arg_strings[] = { var_strings, NULL, NULL };
static const arg_symbol_t  *var_arg_symbols[] = { var_symbols, NULL, NULL };

static const arg_type_t     request_arg_types[] = { CONF_STRING_LIST };
static const char         **request_arg_strings[] = { NULL };
static const arg_symbol_t  *request_arg_symbols[] = { NULL };

static const arg_type_t     require_arg_types[] = { CONF_STRING_LIST };
static const char         **require_arg_strings[] = { NULL };
static const arg_symbol_t  *require_arg_symbols[] = { NULL };

static const arg_type_t     configure_arg_types[] = { CONF_STRING_LIST };
static const char         **configure_arg_strings[] = { NULL };
static const arg_symbol_t  *configure_arg_symbols[] = { NULL };

static const arg_type_t     append_arg_types[] = { CONF_STRING, CONF_ASSIGNMENT, CONF_STRING_LIST };
static const char         **append_arg_strings[] = { NULL, NULL, NULL };
static const arg_symbol_t  *append_arg_symbols[] = { NULL, NULL, NULL };

static const arg_type_t     prepend_arg_types[] = { CONF_STRING, CONF_ASSIGNMENT, CONF_STRING_LIST };
static const char         **prepend_arg_strings[] = { NULL, NULL, NULL };
static const arg_symbol_t  *prepend_arg_symbols[] = { NULL, NULL, NULL };

static const arg_type_t     override_arg_types[] = { CONF_STRING, CONF_ASSIGNMENT, CONF_STRING_LIST };
static const char         **override_arg_strings[] = { NULL, NULL, NULL };
static const arg_symbol_t  *override_arg_symbols[] = { NULL, NULL, NULL };

static const arg_type_t     server_arg_types[] = { CONF_IDENTIFIER, CONF_ADDRESS, CONF_GROUP };
static const char         **server_arg_strings[] = { server_strings, NULL, NULL };
static const arg_symbol_t  *server_arg_symbols[] = { server_symbols, NULL, NULL };

/* command structures. */

/* set command. */
static const command_t command_set = {
    DIRECTIVE_SET,
    "set",
    3,
    var_arg_strings,
    var_arg_types,
    var_arg_symbols,
};

/* set command with boolean. */
static const command_t command_set_boolean = {
    DIRECTIVE_SET_BOOLEAN,
    "enable",
    3,
    var_boolean_arg_strings,
    var_boolean_arg_types,
    var_boolean_arg_symbols,
};

/* request command. */
static const command_t command_request = {
    DIRECTIVE_REQUEST,
    "request",
    1,
    request_arg_strings,
    request_arg_types,
    request_arg_symbols,
};

/* request command. */
static const command_t command_require = {
    DIRECTIVE_REQUIRE,
    "require",
    1,
    require_arg_strings,
    require_arg_types,
    require_arg_symbols,
};

/* configure command. */
static const command_t command_configure = {
    DIRECTIVE_CONFIGURE,
    "configure",
    1,
    configure_arg_strings,
    configure_arg_types,
    configure_arg_symbols,
};

/* server command. */
static const command_t command_server = {
    DIRECTIVE_SERVER_GROUP,
    "server",
    3,
    server_arg_strings,
    server_arg_types,
    server_arg_symbols,
};

/* append command. */
static const command_t command_append = {
    DIRECTIVE_APPEND,
    "append",
    3,
    append_arg_strings,
    append_arg_types,
    append_arg_symbols,
};

/* prepend command. */
static const command_t command_prepend = {
    DIRECTIVE_PREPEND,
    "prepend",
    3,
    prepend_arg_strings,
    prepend_arg_types,
    prepend_arg_symbols,
};

/* prepend command. */
static const command_t command_override = {
    DIRECTIVE_OVERRIDE,
    "override",
    3,
    override_arg_strings,
    override_arg_types,
    override_arg_symbols,
};

static const command_t *commands[] = {
    &command_set_boolean,
    &command_set,
    &command_request,
    &command_require,
    &command_server,
    &command_configure,
    &command_append,
    &command_prepend,
    &command_override,
    NULL,
};

/* forward declaration of directive handlers. */

static int directive_group_handler(client_conf_params_t *params, void *directive_data, directive_group_t group_type, void *group_data);
static int directive_set_handler(client_conf_params_t *params, void *directive_data, directive_group_t group_type, void *group_data);
static int directive_set_boolean_handler(client_conf_params_t *params, void *directive_data, directive_group_t group_type, void *group_data);
static int directive_request_handler(client_conf_params_t *params, void *directive_data, directive_group_t group_type, void *group_data);
static int directive_require_handler(client_conf_params_t *params, void *directive_data, directive_group_t group_type, void *group_data);
static int directive_server_handler(client_conf_params_t *params, void *directive_data, directive_group_t group_type, void *group_data);
static int directive_configure_handler(client_conf_params_t *params, void *directive_data, directive_group_t group_type, void *group_data);
static int directive_append_handler(client_conf_params_t *params, void *directive_data, directive_group_t group_type, void *group_data);
static int directive_prepend_handler(client_conf_params_t *params, void *directive_data, directive_group_t group_type, void *group_data);
static int directive_override_handler(client_conf_params_t *params, void *directive_data, directive_group_t group_type, void *group_data);

/* indexed by directive types in dhcp-client-conf.h . */
directive_handler_t directive_handlers[] = {
    directive_set_boolean_handler,
    directive_set_handler,
    directive_request_handler,
    directive_require_handler,
    directive_server_handler,
    directive_configure_handler,
    directive_append_handler,
    directive_prepend_handler,
    directive_override_handler,
};

/* Basic network configuration.  Here we keep the options we want
 * by default.  We always want these and will override the config
 * passed us by the user. */

static const int default_config_options[] = {
    TAG_DHCP_SUBNET_MASK,           /* Subnet Mask */
    TAG_DHCP_ROUTER,                /* Routers. */
    TAG_DHCP_INTERFACE_MTU,         /* Interface MTU */
    TAG_DHCP_REQUESTED_IP_ADDRESS,  /* Requested IP Address. */
    TAG_DHCP_IP_ADDRESS_LEASE_TIME, /* IP Address lease time. */
    TAG_DHCP_RENEWAL_TIME,          /* Renewal time. */
    TAG_DHCP_REBINDING_TIME,        /* Rebinding time. */
    TAG_DHCP_SERVER_IDENTIFIER,     /* Server Identifier. */
};

/* * * * * * * * * * *
 * utility routines  *
 * * * * * * * * * * */

static void configure_default_request_options(client_conf_t *cc)
{
    conf_var_t *var;
    uint8_t *default_request_options;
    int i;

    default_request_options = xcalloc(MAX_OPTIONS_HANDLED_NUM * sizeof(uint8_t));

    for(i = 0; i < NELMS(default_config_options); i++) {
        default_request_options[default_config_options[i]] = 1;
    }
 
    var = dhcp_conf_var_create(CLIENT_VAR_REQUEST_OPTIONS, default_request_options);
    list_add(cc->params->variables, var);

    return;
}

/* merge two bit arrays into dest. */
static void merge_bit_arrays(uint8_t *dest, uint8_t *src, int len)
{
    int i;

    for(i = 0; i < len; i++) {
        if(src[i])
            dest[i] = 1;
    }

    return;
}

/* search through a list of conf variables and return if the type
 * is found.  this is useful for overwriting variables: this is
 * slow however we assume we're not doing configuration variables
 * in excess of twenty or so in most situations. */
static conf_var_t *find_conf_var(var_type_t var_type, list_t *var_list)
{
    conf_var_t *var;

    list_rewind(var_list);
    while((var = list_next(var_list)) != NULL)  {

        if(var->type == var_type)
            return var;
    }

    return NULL;
}

/* add variable or replace existing one. */
static void add_or_replace_var(list_t *variables, conf_var_t *var)
{
    conf_var_t *cur_var;

    /* see if this is already set.*/
    cur_var = find_conf_var(var->type, variables);

    /* if not add. */
    if(cur_var == NULL) {

        list_add(variables, var);

    } else {

        /* otherwise overwrite. */
        dhcp_conf_var_overwrite(cur_var, var->type, var->val);
        dhcp_conf_var_destroy_no_val_free(var); /* we need to keep the value so don't free val. */

    }
}

/* add or overwrite the bit array option. */
static void add_or_overwrite_bit_array_var(list_t *variables, uint8_t *new_array, var_type_t var_type)
{
    conf_var_t *cur_var;
    uint8_t *old_array;

    /* see if this is already set.*/
    cur_var = find_conf_var(var_type, variables);

    /* if not add. */
    if(cur_var == NULL) {

        cur_var = dhcp_conf_var_create(var_type, new_array);
        list_add(variables, cur_var);

    } else {

        old_array = dhcp_conf_var_get_val(cur_var);
        merge_bit_arrays(old_array, new_array, MAX_OPTIONS_HANDLED_NUM);
        xfree(new_array); /* free up since we're not replacing with it. */

    }

    return;
}

/* search through lower params and return the first match. */
static client_conf_params_t *get_params_for(list_t *lower_params, ip_addr_t ip, eth_addr_t eth_addr)
{
    client_conf_params_t *params = NULL;

    list_rewind(lower_params);
    while((params = list_next(lower_params)) != NULL) {

        if((memcmp(&ip, &params->server_ip, sizeof(ip_addr_t)) == 0) 
           || (memcmp(&eth_addr, &params->server_hw_addr, sizeof(eth_addr_t)) == 0)) {
            /* if a match is found break out. */
            break;
        } else {
            /* check lower parameters too. */
            if((params = get_params_for(params->lower_params, ip, eth_addr)) != NULL)
                break;
        }

        /* otherwise we keep looping. */

    }

    return params;
}

/* utility routine to call on client_conf to find the best
 * param. this calls get_params_for on all lower parameters which is in of itself recursive. */
static client_conf_params_t *find_best_params_for(client_conf_t *cc, ip_addr_t ip_addr, eth_addr_t eth_addr)
{
    client_conf_params_t *params;

    params = get_params_for(cc->params->lower_params, ip_addr, eth_addr);
    if(params == NULL)
        params = cc->params;
    
    return params;
}

/* given a list of strings identifying options we fill a bit array that references the option by the index. */
static int client_conf_fill_option_bit_array_from_string_list(list_t *strings, uint8_t *option_bit_array)
{
    char *option_string;
    int i;
    const char **option_strings = dhcp_option_conf_string_get_array();

    list_rewind(strings);
    while((option_string = list_next(strings)) != NULL) {

        for(i = 0; i < MAX_OPTIONS_HANDLED; i++) {

            if(!strcmp(option_strings[i], option_string)) {
                option_bit_array[i] = 1;
                break;
            }
        }

        if(i > MAX_OPTIONS_HANDLED) {
            ERROR_MESSAGE("specified unknown option: %s", option_string);
            return 1;
        }
    }

    return 0;
}

/* setup a variable in params_t. */
static int client_conf_set_variable(client_conf_params_t *params, arg_symbol_t var_symbol, const char *var_value)
{
    conf_var_t *var = NULL;
    int *int_val;
    uint16_t *uint16_val;

    switch(var_symbol) {

        /* string types can be created with no checking.
         * just attempt a copy (will exit on failure). */

    case CLIENT_VAR_HOSTNAME:
    case CLIENT_VAR_SUBNET_MASK:
        var = dhcp_conf_var_create(var_symbol, xstrdup(var_value));
        break;

        /* integer values greater than zero. use unsigned check
         * to prevent negative values. */

    case CLIENT_VAR_DHCP_DISCOVER_TIMEOUT:
    case CLIENT_VAR_DHCP_REQUEST_TIMEOUT:
    case CLIENT_VAR_ICMP_ECHO_TIMEOUT:
    case CLIENT_VAR_ICMP_SUBNET_TIMEOUT:
    case CLIENT_VAR_ARP_TIMEOUT:
    case CLIENT_VAR_DHCP_DISCOVER_RETRIES:
    case CLIENT_VAR_DHCP_REQUEST_RETRIES:
    case CLIENT_VAR_ARP_RETRIES:
    case CLIENT_VAR_ICMP_RETRIES:

        if(!is_signed_numeric(var_value))  {
            ERROR_MESSAGE("value not unsigned numeric as expected : %s\n", var_value);
            return 1;
        }

        int_val = xmalloc(sizeof(int));
        *int_val = atoi(var_value);

        var = dhcp_conf_var_create(var_symbol, int_val);
        break;

    case CLIENT_VAR_RENEW_PERCENT:
    case CLIENT_VAR_REBIND_PERCENT:
    case CLIENT_VAR_INTERFACE_MTU:

        uint16_val = xmalloc(sizeof(uint16_t));
        sscanf(var_value, "%"SCNu16, uint16_val);

        if(uint16_val != NULL)  {
            var = dhcp_conf_var_create(var_symbol, uint16_val);
            break;
        }

        /* fall through. */
    default:

        ERROR_MESSAGE("unable to understand assignment value of : %s\n", var_value);
        return 1;
    }

    add_or_replace_var(params->variables, var);
    return 0;
}

/* setup a boolean variable. */
static int client_conf_set_variable_boolean(client_conf_params_t *params, arg_symbol_t var_symbol, uint8_t bool_val)
{
    conf_var_t *var = NULL;
    uint8_t *val;

    val = xmalloc(sizeof(uint8_t));
    *val = bool_val;

    var = dhcp_conf_var_create(var_symbol, val);
    add_or_replace_var(params->variables, var);

    return 0;
}

static char *get_conf_options_fname(client_conf_t *cc)
{
    char *fname;
    stringbuffer_t *sb;

    sb = stringbuffer_create();
    stringbuffer_aprintf(sb, "%s/default.conf", DHCPSYSCONF_CLIENTDIR);

    if(file_exists(stringbuffer_getstring(sb))) {

        fname = xstrdup(stringbuffer_getstring(sb));
        stringbuffer_destroy(sb);
        return fname;

    }

    stringbuffer_clear(sb);
    stringbuffer_aprintf(sb, "%s/%s.conf", DHCPSYSCONF_CLIENTDIR, cc->interface);

    if(file_exists(stringbuffer_getstring(sb))) {

        fname = xstrdup(stringbuffer_getstring(sb));
        stringbuffer_destroy(sb);
        return fname;

    }

    /* otherwise there is no configuration file and this is
     * bad. signal a fatal error. */

    FATAL_MESSAGE("configuration file missing -- looked for default.conf and <interface>.conf. exiting...");
    exit(1); /* get rid of compiler warning. */
}

static uint8_t *get_conf_bit_array_var(client_conf_params_t *params, var_type_t var_type)
{
    conf_var_t *var;
    uint8_t *bit_array = NULL;

    var = find_conf_var(var_type, params->variables);

    if(var != NULL)
        bit_array = dhcp_conf_var_get_val(var);

    return bit_array;
}

static int get_conf_int_var(client_conf_params_t *params, var_type_t var_type, int default_val)
{
    conf_var_t *var;
    int *retries = NULL;

    var = find_conf_var(var_type, params->variables);

    if(var != NULL)
        retries = dhcp_conf_var_get_val(var);

    if(retries)
        return *retries;
    else
        return default_val;
}

static uint8_t get_conf_bool_val(client_conf_params_t *params, var_type_t var_type, uint8_t default_val)
{
    conf_var_t *var;
    uint8_t *bool_val = NULL;

    var = find_conf_var(var_type, params->variables);

    if(var != NULL)
        bool_val = dhcp_conf_var_get_val(var);

    if(bool_val)
        return *bool_val;
    else
        return default_val;
}

static uint16_t get_conf_uint16_val(client_conf_params_t *params, var_type_t var_type, uint16_t default_val)
{
    conf_var_t *var;
    uint16_t *uint16_val = NULL;

    var = find_conf_var(var_type, params->variables);

    if(var != NULL)
        uint16_val = dhcp_conf_var_get_val(var);

    if(uint16_val)
        return *uint16_val;
    else
        return default_val;
}

static const char *get_conf_string_val(client_conf_params_t *params, var_type_t var_type, const char *default_val)
{
    conf_var_t *var;
    const char *string_val = NULL;

    var = find_conf_var(var_type, params->variables);

    if(var != NULL)
        string_val = dhcp_conf_var_get_val(var);

    if(string_val)
        return string_val;
    else
        return default_val;
}

static int client_conf_load_options(client_conf_t *cc)
{
    conf_t *conf;
    list_t *directives;

    conf = conf_create(commands, cc->conf_file);

    if(conf_compile_directives(conf) == CONF_ERROR) {
        ERROR_MESSAGE("error reading conf file: %s at line %d", cc->conf_file, conf_get_line_no(conf));
        conf_destroy(conf);
        return 1;
    }

    directives = conf_get_directives(conf);

    /* read compiled directives. */

    if(directive_group_handler(cc->params, conf->directives, DIRECTIVE_GROUP_NULL, NULL)) {
        conf_destroy(conf);
        return 1;
    }

    conf_destroy(conf);
    return 0;
}

/* * * * * * * * * * * * * * * * * * *
 * params constructors/destructors   *
 * * * * * * * * * * * * * * * * * * */

static client_conf_params_t *params_create(void)
{
    client_conf_params_t *params;

    params = xmalloc(sizeof(client_conf_params_t));
    params->group_type = DIRECTIVE_GROUP_NULL;
    params->variables = list_create();
    params->lower_params = list_create();
    params->append_options = list_create();
    params->prepend_options = list_create();
    params->override_options = list_create();

    return params;
}

static void params_destroy_l(void *data)
{
    client_conf_params_t *params = data;

    list_destroy(params->variables, dhcp_conf_var_destroy_l);
    list_destroy(params->lower_params, params_destroy_l);

    dhcp_opt_destroy_option_list(params->append_options);
    dhcp_opt_destroy_option_list(params->prepend_options);
    dhcp_opt_destroy_option_list(params->override_options);

    xfree(params);
}


static void params_destroy(client_conf_params_t *params)
{
    /* call recursive function to destroy it and everything under it. */
    params_destroy_l(params);
    return;
}

/* * * * * * *
 * accessors *
 * * * * * * */

/* get options to override. */
list_t *client_conf_get_override_options(client_conf_t *cc, ip_addr_t ip_addr, eth_addr_t eth_addr)
{
    return cc->params->override_options;
}

/* get options to append. */
list_t *client_conf_get_append_options(client_conf_t *cc, ip_addr_t ip_addr, eth_addr_t eth_addr)
{
    return cc->params->append_options;
}

/* get options to prepend. */
list_t *client_conf_get_prepend_options(client_conf_t *cc, ip_addr_t ip_addr, eth_addr_t eth_addr)
{
    return cc->params->prepend_options;
}

/* get parameters to request: this is always from the top level. */
uint8_t *client_conf_get_request_opt_bit_array(client_conf_t *cc)
{
    return get_conf_bit_array_var(cc->params, CLIENT_VAR_REQUEST_OPTIONS);
}

/* get number of retries for dhcp: this is always from the top level. */
int client_conf_get_dhcp_discovery_retries(client_conf_t *cc)
{
    return get_conf_int_var(cc->params, CLIENT_VAR_DHCP_DISCOVER_RETRIES, CLIENT_DEFAULT_DHCP_DISCOVER_RETRIES);
}

/* get number of retries for dhcp: this is always from the top level. */
int client_conf_get_dhcp_request_retries(client_conf_t *cc)
{
    return get_conf_int_var(cc->params, CLIENT_VAR_DHCP_REQUEST_RETRIES, CLIENT_DEFAULT_DHCP_REQUEST_RETRIES);
}

/* get required options */
uint8_t *client_conf_get_opt_required_bit_array(client_conf_t *cc, ip_addr_t ip_addr, eth_addr_t eth_addr)
{
    client_conf_params_t *params = find_best_params_for(cc, ip_addr, eth_addr);
    return get_conf_bit_array_var(params, CLIENT_VAR_REQUIRE_OPTIONS);
}

/* get configure options */
uint8_t *client_conf_get_opt_configure_bit_array(client_conf_t *cc, ip_addr_t ip_addr, eth_addr_t eth_addr)
{
    client_conf_params_t *params = find_best_params_for(cc, ip_addr, eth_addr);
    return get_conf_bit_array_var(params, CLIENT_VAR_CONFIGURE_OPTIONS);
}

/* get number of times to attempt icmp operations where applicable */
int client_conf_get_icmp_retries(client_conf_t *cc, ip_addr_t ip_addr, eth_addr_t eth_addr)
{
    client_conf_params_t *params = find_best_params_for(cc, ip_addr, eth_addr);
    return get_conf_int_var(params, CLIENT_VAR_ICMP_RETRIES, CLIENT_DEFAULT_ICMP_RETRIES);
}

/* get number of times to attempt arp operations where applicable */
int client_conf_get_arp_retries(client_conf_t *cc, ip_addr_t ip_addr, eth_addr_t eth_addr)
{
    client_conf_params_t *params = find_best_params_for(cc, ip_addr, eth_addr);
    return get_conf_int_var(params, CLIENT_VAR_ARP_RETRIES, CLIENT_DEFAULT_ARP_RETRIES);
}

/* get timeout threshold for dhcp discover operations. */
int client_conf_get_dhcp_discover_timeout(client_conf_t *cc)
{
    return get_conf_int_var(cc->params, CLIENT_VAR_DHCP_DISCOVER_TIMEOUT, CLIENT_DEFAULT_DHCP_DISCOVER_TIMEOUT);
}

/* get timeout threshold for dhcp request operations. */
int client_conf_get_dhcp_request_timeout(client_conf_t *cc)
{
    return get_conf_int_var(cc->params, CLIENT_VAR_DHCP_REQUEST_TIMEOUT, CLIENT_DEFAULT_DHCP_REQUEST_TIMEOUT);
}

/* get timeout threshold for arp operations. */
int client_conf_get_arp_timeout(client_conf_t *cc, ip_addr_t ip_addr, eth_addr_t eth_addr)
{
    client_conf_params_t *params = find_best_params_for(cc, ip_addr, eth_addr);
    return get_conf_int_var(params, CLIENT_VAR_ARP_TIMEOUT, CLIENT_DEFAULT_ARP_TIMEOUT);
}

/* get timeout threshold for icmp echo operations. */
int client_conf_get_icmp_echo_timeout(client_conf_t *cc, ip_addr_t ip_addr, eth_addr_t eth_addr)
{
    client_conf_params_t *params = find_best_params_for(cc, ip_addr, eth_addr);
    return get_conf_int_var(params, CLIENT_VAR_ICMP_ECHO_TIMEOUT, CLIENT_DEFAULT_ICMP_ECHO_TIMEOUT);
}

/* get timeout threshold for icmp subnet operations. */
int client_conf_get_icmp_subnet_timeout(client_conf_t *cc, ip_addr_t ip_addr, eth_addr_t eth_addr)
{
    client_conf_params_t *params = find_best_params_for(cc, ip_addr, eth_addr);
    return get_conf_int_var(params, CLIENT_VAR_ICMP_SUBNET_TIMEOUT, CLIENT_DEFAULT_ICMP_SUBNET_TIMEOUT);
}

/* get the hostname option the client should set: always from the top level */
const char *client_conf_get_hostname(client_conf_t *cc)
{
    return get_conf_string_val(cc->params, CLIENT_VAR_HOSTNAME, NULL);
}

/* find out whether or not we should attempt to measure router latency. */
uint8_t client_conf_get_do_measure_router_latency_icmp(client_conf_t *cc, ip_addr_t ip_addr, eth_addr_t eth_addr)
{
   client_conf_params_t *params = find_best_params_for(cc, ip_addr, eth_addr);
   return get_conf_bool_val(params, CLIENT_VAR_DO_MEASURE_ROUTER_LATENCY_ICMP, CLIENT_DEFAULT_DO_MEASURE_LATENCY_ICMP);
}

/* get mtu setting. */
int client_conf_get_default_mtu(client_conf_t *cc)
{
    return get_conf_uint16_val(cc->params, CLIENT_VAR_INTERFACE_MTU, CLIENT_DEFAULT_MTU);
}

const char *client_conf_get_default_subnet_mask(client_conf_t *cc, ip_addr_t ip_addr, eth_addr_t eth_addr)
{
    client_conf_params_t *params = find_best_params_for(cc, ip_addr, eth_addr);
    return get_conf_string_val(params, CLIENT_VAR_SUBNET_MASK, CLIENT_DEFAULT_SUBNET_MASK);
}

int client_conf_get_renew_percent(client_conf_t *cc)
{
    return get_conf_uint16_val(cc->params, CLIENT_VAR_RENEW_PERCENT, CLIENT_DEFAULT_RENEW_PERCENT);
}

int client_conf_get_rebind_percent(client_conf_t *cc)
{
    return get_conf_uint16_val(cc->params, CLIENT_VAR_REBIND_PERCENT, CLIENT_DEFAULT_REBIND_PERCENT);
}

/* * * * * * * * * * * *
 * directive handlers. *
 * * * * * * * * * * * */

static int directive_group_handler(client_conf_params_t *params, void *directive_data, directive_group_t group_type, void *group_data)
{
    list_t *directive_list;
    directive_t *directive;

    /* setup group before handling directives under it. */
    params->group_type = group_type;

    /* copy over data if applicable. */
    switch(group_type) {

    case DIRECTIVE_GROUP_IP_ADDRESS:

        memcpy(&params->server_ip, group_data, IP_ADDR_LEN);
        break;

    case DIRECTIVE_GROUP_MAC_ADDRESS:

        memcpy(&params->server_hw_addr, group_data, ETH_ADDR_LEN);
        break;

    case DIRECTIVE_GROUP_NULL:  /* fall through. */
    default:
        break;

    }

    /* now process group. */
    directive_list = directive_data;
    list_rewind(directive_list);
    while((directive = list_next(directive_list)) != NULL) {

        if(directive_handlers[directive->command_code](params, directive, group_type, group_data))
            return 1;
    }

    return 0;
}

static int directive_set_handler(client_conf_params_t *params, void *directive_data, directive_group_t group_type, void *group_data)
{
    directive_t *directive = directive_data;
    list_t *args;
    arg_symbol_t *var_symbol;
    char *var_value;

    args = directive->arguments;

    var_symbol = list_get_by_index(args, 0);
    var_value = list_get_by_index(args, 2);

    if(client_conf_set_variable(params, *var_symbol, var_value)) {
        ERROR_MESSAGE("configuration: error unable to set variable %s to %s", var_strings[*var_symbol], var_value);
        return 1;
    }

    return 0;
}

/* procified routine for request/require/configure etc: these all deal with bit arrays. */
static int directive_option_array_handler_proc(client_conf_params_t *params, void *directive_data, 
                                               directive_group_t group_type, void *group_data, 
                                               var_type_t type)
{
    directive_t *directive = directive_data;
    list_t *args;
    list_t *string_list;
    uint8_t *options;

    /* the first argument is the string list we want. */
    args = directive->arguments;
    string_list = list_first(args);

    /* create bit array. */
    options = xcalloc(MAX_OPTIONS_HANDLED_NUM * sizeof(uint8_t));

    if(client_conf_fill_option_bit_array_from_string_list(string_list, options)) {
        xfree(options);
        return 1;
    }

    add_or_overwrite_bit_array_var(params->variables, options, type);
    return 0;

}

static int directive_request_handler(client_conf_params_t *params, void *directive_data, directive_group_t group_type, void *group_data)
{
    return directive_option_array_handler_proc(params, directive_data, 
                                               group_type, group_data, 
                                               CLIENT_VAR_REQUEST_OPTIONS);
}

static int directive_require_handler(client_conf_params_t *params, void *directive_data, directive_group_t group_type, void *group_data)
{
    return directive_option_array_handler_proc(params, directive_data, 
                                               group_type, group_data, 
                                               CLIENT_VAR_REQUIRE_OPTIONS);
}

static int directive_configure_handler(client_conf_params_t *params, void *directive_data, directive_group_t group_type, void *group_data)
{
    return directive_option_array_handler_proc(params, directive_data, 
                                               group_type, group_data, 
                                               CLIENT_VAR_CONFIGURE_OPTIONS);
}

static int directive_user_defined_override(client_conf_params_t *params, const char *option_name, dhcp_opt_t *option)
{
    if(option == NULL) {
        ERROR_MESSAGE("unknown option %s to override", option_name);
        return 1;
    }

    list_add(params->override_options, option);
    return 0;
}

static int directive_user_defined_append(client_conf_params_t *params, const char *option_name, dhcp_opt_t *option)
{
    if(option == NULL) {
        ERROR_MESSAGE("unknown option %s to append", option_name);
        return 1;
    }

    /* cannot append unless it's a list type. */

    if(dhcp_opt_get_type(option) != DHCP_OPT_LIST)  {
        ERROR_MESSAGE("user config attempted to append to: %s", option_name);
        ERROR_MESSAGE("cannot append option which is not a list type");
        ERROR_MESSAGE("you can only override options which are not lists.");

        dhcp_opt_destroy(option);
        return 1;
    }

    list_add(params->append_options, option);
    return 0;
}

static int directive_user_defined_prepend(client_conf_params_t *params, const char *option_name, dhcp_opt_t *option)
{
    if(option == NULL) {
        ERROR_MESSAGE("unknown option %s to prepend", option_name);
        return 1;
    }
    /* cannot append unless it's a list type. */

    if(dhcp_opt_get_type(option) != DHCP_OPT_LIST)  {
        ERROR_MESSAGE("user config attempted to prepend to: %s", option_name);
        ERROR_MESSAGE("cannot prepend option which is not a list type");
        ERROR_MESSAGE("you can only override options which are not lists.");

        dhcp_opt_destroy(option);
        return 1;
    }

    list_add(params->prepend_options, option);
    return 0;
}

static int directive_user_defined_option_handler_proc(client_conf_params_t *params, void *directive_data, 
                                                      directive_group_t group_type, void *group_data,
                                                      int var_type)
{
    char *option_name;
    list_t *option_data_list;
    dhcp_opt_t *option = NULL;
    directive_t *directive = directive_data;
    uint8_t i;

    const char **option_strings = dhcp_option_conf_string_get_array();

    option_name = list_first(directive->arguments);
    option_data_list = list_get_by_index(directive->arguments, 2);

    for(i = 0; i < MAX_OPTIONS_HANDLED; i++) {

        if(!strcmp(option_strings[i], option_name)) {

            option = dhcp_opt_create_from_user_string(i, option_data_list);
            break;
        }
    }

    if(i == MAX_OPTIONS_HANDLED) {
        ERROR_MESSAGE("unrecognized dhcp option specified by user config: %s", option_name);
        return 1;
    }

    /* check to make sure option name is valid. */
    
    switch(var_type) {

    case CLIENT_VAR_APPEND_OPTIONS:
        directive_user_defined_append(params, option_name, option);
        break;

    case CLIENT_VAR_PREPEND_OPTIONS:
        directive_user_defined_prepend(params, option_name, option);
        break;

    case CLIENT_VAR_OVERRIDE_OPTIONS:
        directive_user_defined_override(params, option_name, option);
        break;

    default:
        FATAL_MESSAGE("illegal append/prepend/override specified. i have encountered a bug. i shouldn't be here.");
    }

    return 0;
}

static int directive_append_handler(client_conf_params_t *params, void *directive_data, directive_group_t group_type, void *group_data)
{
    return directive_user_defined_option_handler_proc(params, directive_data, group_type, group_data, CLIENT_VAR_APPEND_OPTIONS);
}

static int directive_prepend_handler(client_conf_params_t *params, void *directive_data, directive_group_t group_type, void *group_data)
{
    return directive_user_defined_option_handler_proc(params, directive_data, group_type, group_data, CLIENT_VAR_PREPEND_OPTIONS);
}

static int directive_override_handler(client_conf_params_t *params, void *directive_data, directive_group_t group_type, void *group_data)
{
    return directive_user_defined_option_handler_proc(params, directive_data, group_type, group_data, CLIENT_VAR_OVERRIDE_OPTIONS);
}

static int directive_server_handler(client_conf_params_t *params, void *directive_data, directive_group_t group_type, void *group_data)
{
    list_t *args;
    const arg_symbol_t *server_symbol;
    void *address_data;
    void *directives;
    client_conf_params_t *new_params;
    directive_group_t new_group_type;
    directive_t *directive;

    directive = directive_data;
    args = directive->arguments;
    server_symbol = list_first(args);

    switch(*server_symbol) {

    case CLIENT_SERVER_IP:
        /* ip address. */

        new_group_type = DIRECTIVE_GROUP_IP_ADDRESS;
        break;


    case CLIENT_SERVER_MAC:
        /* mac address. */

        new_group_type = DIRECTIVE_GROUP_MAC_ADDRESS;
        break;

    default:
        FATAL_MESSAGE("illegal server directive group type. this is a bug report me.");
        exit(1); /* get rid of pesky compiler warnings. */
    }

    address_data = list_get_by_index(args, 1);
    directives = list_get_by_index(args, 2);

    new_params = params_create();
    if(directive_group_handler(new_params, directives, new_group_type, address_data) < 0)
        return 1;

    list_add_to_end(params->lower_params, new_params);

    return 0;
}

static int directive_set_boolean_handler(client_conf_params_t *params, void *directive_data, 
                                         directive_group_t group_type, void *group_data)
{
    directive_t *directive = directive_data;
    list_t *args;
    arg_symbol_t *var_symbol;
    uint8_t *var_value;

    args = directive->arguments;

    var_symbol = list_get_by_index(args, 0);
    var_value = list_get_by_index(args, 2);

    if(client_conf_set_variable_boolean(params, *var_symbol, *var_value)) {
        ERROR_MESSAGE("configuration: error unable to set boolean variable %s", var_strings[*var_symbol], var_value);
        return 1;
    }

    return 0;
}

client_conf_t *create_client_conf(const char *interface)
{
    client_conf_t *cc;

    cc = xcalloc(sizeof(client_conf_t));
    cc->interface = interface;
    cc->conf_file = get_conf_options_fname(cc);
    cc->params = params_create(); /* set top level to catchall group. */
    cc->params->group_type = DIRECTIVE_GROUP_NULL;

    /* setup defaults request options: these we always request */
    configure_default_request_options(cc);

    if(client_conf_load_options(cc)) {
        client_conf_destroy(cc);
        return NULL;
    }

    return cc;
}

void client_conf_destroy(client_conf_t *cc)
{
    /* destroy loaded parameters. */

    params_destroy(cc->params);

    /* free up the client conf structure. */
    xfree(cc->conf_file);
    xfree(cc);

    return;

}

int client_conf_reread(client_conf_t *cc)
{
    if(client_conf_load_options(cc)){
        return 1;
    }

    return 0;
}
