/* $Header: /cvsroot/dhcp-agent/dhcp-agent/src/dhcp-client-states.c,v 1.44 2003/06/28 17:50:06 actmodern Exp $
 * 
 * Copyright 2002 Thamer Alharbash <tmh@whitefang.com>
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in
 * the documentation and/or other materials provided with the
 * distribution.
 * 3. The names of the authors may not be used to endorse or promote
 * products derived from this software without specific prior
 * written permission.
 * 
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 * 
 * These are the individual client state functions, 
 * along with their utility routines.
 * 
 */

#define MODULE_NAME "dhcp-client-states"

#include "dhcp-local.h"
#include "dhcp-limits.h"
#include "dhcp-libutil.h"
#include "dhcp-librawnet.h"
#include "dhcp-option.h"

#include "dhcp-options-strings.h"
#include "dhcp-client-conf.h"
#include "dhcp-client-cache.h"
#include "dhcp-client.h"
#include "dhcp-sysconf.h"


/***************************
 * option list generation  *
 ***************************/

/* create DISCOVER and REQUEST option list. */
static list_t *client_build_option_list(dhcp_client_control_t *dc)
{
    list_t *options;
    dhcp_opt_t *option;
    char *hostname_tmp;
    const char *hostname;
    uint16_t max_message_opt_data = MAX_MESSAGE_SIZE;

    options = list_create();

    option = dhcp_opt_create_from_internal_data(TAG_DHCP_MAX_DHCP_SIZE, &max_message_opt_data, 
                                                sizeof(max_message_opt_data));
    list_add(options, option);

    /* client_id */
    option = dhcp_opt_create_from_internal_data(TAG_DHCP_CLIENT_ID, dc->client_id, (ETH_ADDR_LEN + 1));
    list_add(options, option);
    /* class_id */
    option = dhcp_opt_create_from_internal_data(TAG_DHCP_VENDOR_CLASS_ID, dc->class_id, (strlen(dc->class_id) + 1));
    list_add(options, option);

    /* hostname if available */
    hostname = client_conf_get_hostname(dc->conf);
    if(hostname != NULL) {
        /* if we have one set it. */

        /* can't pass const char pointer so copy to char pointer.
         * FIXME: audit types and fix any const propogation
         * problems like these properly. */

        hostname_tmp = xstrdup(hostname);
        option = dhcp_opt_create_from_internal_data(TAG_DHCP_HOST_NAME, hostname_tmp, (strlen(hostname_tmp) + 1));
        xfree(hostname_tmp);

        list_add(options, option);
    }

    return options;
}

/* create DHCP DISCOVER option list. */
static list_t *client_build_discover_option_list(dhcp_client_control_t *dc)
{
    list_t *options;
    list_t *requested_options;
    dhcp_opt_t *option;

    options = list_create();
    option = dhcp_opt_create_parameter_request_list(client_conf_get_request_opt_bit_array(dc->conf));

    if(option == NULL) {
        WARN_MESSAGE("sending without parameter request list");
    } else {
        list_add(options, option);
    }

    /* dhcp message type. */
    option = dhcp_opt_create_message_type(DHCP_DISCOVER_TM);
    list_add(options, option);

    requested_options = client_build_option_list(dc);
    list_join(options, requested_options); /* this destroys requested_options */

    return options;
}

/* create DHCP REQUEST option list. */
static list_t *client_build_request_option_list(dhcp_client_control_t *dc)
{
    list_t *options;
    list_t *requested_options;
    dhcp_opt_t *option;

    options = list_create();

    /* dhcp message type. */
    option = dhcp_opt_create_message_type(DHCP_REQUEST_TM);
    list_add(options, option);

    requested_options = client_build_option_list(dc);
    list_join(options, requested_options); /* this destroys requested_options */

    return options;
}

/* create DHCP RELEASE option list. */
static list_t *client_build_release_option_list(dhcp_client_control_t *dc)
{
    list_t *options;
    dhcp_opt_t *option;

    options = list_create();

    /* client_id: RFC2131 says we MAY */
    option = dhcp_opt_create_from_internal_data(TAG_DHCP_CLIENT_ID, dc->client_id, (ETH_ADDR_LEN + 1));
    list_add(options, option);

    /* dhcp message type. */
    option = dhcp_opt_create_message_type(DHCP_RELEASE_TM);
    list_add(options, option);

    return options;

}

static list_t *client_build_decline_option_list(dhcp_client_control_t *dc, ip_addr_t ip_address)
{
    list_t *options;
    dhcp_opt_t *option;

    options = list_create();

    /* client_id: RFC2131 says we MAY */
    option = dhcp_opt_create_from_internal_data(TAG_DHCP_CLIENT_ID, dc->client_id, (ETH_ADDR_LEN + 1));
    list_add(options, option);

    /* requested ip address: RFC2131 MUST. */
    option = dhcp_opt_create_from_internal_data(TAG_DHCP_REQUESTED_IP_ADDRESS, &ip_address, IP_ADDR_LEN);
    list_add(options, option);

    /* dhcp message type. */
    option = dhcp_opt_create_message_type(DHCP_DECLINE_TM);
    list_add(options, option);

    return options;
}

/* create DHCP RENEW option list. */
static list_t *client_build_renew_option_list(dhcp_client_control_t *dc)
{
    list_t *options;
    dhcp_opt_t *option;

    options = list_create();

    /* dhcp message type. */
    option = dhcp_opt_create_message_type(DHCP_REQUEST_TM);
    list_add(options, option);

    /* client_id */
    option = dhcp_opt_create_from_internal_data(TAG_DHCP_CLIENT_ID, dc->client_id, (ETH_ADDR_LEN + 1));
    list_add(options, option);

    return options;
}

/* create DHCP REBIND option list. */
static list_t *client_build_rebind_option_list(dhcp_client_control_t *dc)
{
    /* just call renew_option_list since it's the same. */
    return (client_build_renew_option_list(dc));
}

/*************
 * ARP/UNARP *
 *************/

/* 
 * Use this to send out a broadcast reply claiming that we own the
 * IP. To prevent a poisoned cache we use the actual interface hw
 * addr as opposed to any fake one we may have been passed.
 * 
 */

static int client_broadcast_arp_reply(dhcp_client_control_t *dc, ip_addr_t client_ip_addr)
{
    eth_addr_t client_hw_addr;

    client_ip_addr = rawnet_get_ip_addr(dc->rawnet);
    rawnet_get_real_hw_addr(dc->rawnet, &client_hw_addr);

    build_arp_reply_broadcast(dc->rawnet, client_ip_addr, client_hw_addr);
    if(rawnet_send_packet(dc->rawnet) < 0) {
        ERROR_MESSAGE("could not transmit arp reply.");
        return -1;
    }

    return 0;
}

/* UNARP (rfc 1868). This makes us more network friendly. */
static int client_broadcast_unarp(dhcp_client_control_t *dc)
{
    ip_addr_t client_ip_addr;
    eth_addr_t client_hw_addr;

    client_ip_addr = rawnet_get_ip_addr(dc->rawnet);
    rawnet_get_real_hw_addr(dc->rawnet, &client_hw_addr);

    build_unarp(dc->rawnet, client_ip_addr, client_hw_addr);
    if(rawnet_send_packet(dc->rawnet) < 0) {
        ERROR_MESSAGE("could not transmit unarp.");
        return -1;
    }

    return 0;
}

/***************************
 * utility routines        *
 ***************************/

/* update secs field. can be called from rawnet_packet_transact */
static void client_update_packet(void *arg)
{
    dhcp_client_control_t *dc = arg;

    dhcp_client_update_secs(dc);
    rawnet_dhcp_update(dc->rawnet, (uint16_t)dc->secs); /* FIXME: this may be an integer overflow. */
    return;
}

/* setup our timers. */
static void client_setup_timers(dhcp_client_control_t *dc)
{
    list_t *timer_options;

    timer_options = client_cache_load_timer_options(dc->cache, 0);

    if(list_get_len(timer_options) == 0) {
        FATAL_MESSAGE("no timers to setup. we should have least one: lease time");
        exit(1);
    }

    do_sysconf_setup_timers(timer_options, dc);
    dhcp_opt_destroy_option_list(timer_options);

    return;
}

static time_t compensate_timeout_with_timer(dhcp_client_control_t *dc)
{
    uint32_t next_timer;
    time_t timeout;

    /* we realize time_t may not be uint32_t, however we're only
     * going to set the max timeout if it is lower so this is
     * safe. we shouldn't overflow. */

    next_timer = timer_peek_next_timer(dc->context->timer);
    if(next_timer < RECOMMENDED_MAX_SECS_WAIT) {

        if(next_timer != 0)
            timeout = next_timer / 2;
        else
            timeout = next_timer;

    } else {
        timeout = RECOMMENDED_MAX_SECS_WAIT;
    }

   return timeout;
}

/* make a new copy without options which are meant for protocol purposes only. */
static list_t *client_remove_protocol_only_options(list_t *options)
{
    dhcp_opt_t *option, *option_copy;
    list_t *non_protocol_opts;

    non_protocol_opts = list_create();
    list_rewind(options);
    while((option = list_next(options)) != NULL) {

        switch(dhcp_opt_get_tag(option)) {

        case TAG_DHCP_MAX_DHCP_SIZE:
        case TAG_DHCP_MESSAGE_TYPE:
        case TAG_DHCP_OVERLOAD:
            break;

        default:
            option_copy = dhcp_option_copy(option);
            list_add(non_protocol_opts, option_copy);
            break;
        }
    }

    return non_protocol_opts;
}

/* utility routine that should zap our configuration and leave us
 * with an interface with no real address on it. */
static void client_unconfigure(dhcp_client_control_t *dc)
{
    list_t *options;
    ip_addr_t ip_addr;

    /* load our cache and call sysconf with DO_SHUTDOWN to
     * unconfigure us. */

    options = client_cache_load_options(dc->cache, 0);
    do_sysconf(options, dc, STATE_SHUTDOWN);
    dhcp_opt_destroy_option_list(options);

    ip_addr = rawnet_get_ip_addr(dc->rawnet);

    /* check if we have a real address or not. */
    if(ip_addr != 0) {

        /* call UNARP first. */
        client_broadcast_unarp(dc);

        /* zap our ip address. */
        rawnet_interface_up(dc->rawnet, 0, 0, CLIENT_DEFAULT_MTU, 1);
    }

    /* we're done. */
    return;
}

/* this is called from setup: we should have a lease in our cache. */
static int client_configure(dhcp_client_control_t *dc)
{
    list_t *options;
    ip_addr_t interface_ip;
    ip_addr_t *passed_ip = NULL;
    eth_addr_t eth_addr;
    dhcp_opt_t *option;

    INFO_MESSAGE("performing system configuration");

    options = client_cache_load_options(dc->cache, 0);
    if(list_get_len(options) == 0) {
        ERROR_MESSAGE("received empty configuration cache! cannot configure host");
        return STATE_FATAL_ERROR;
    }

    /* Check if the address we've been assigned is in use.  only
     * check if our interface address does not match the address
     * in our lease cache.*/

    interface_ip = rawnet_get_ip_addr(dc->rawnet);

    list_rewind(options);
    while((option = list_next(options)) != NULL) {

        if(dhcp_opt_get_tag(option) == TAG_DHCP_REQUESTED_IP_ADDRESS) {
            passed_ip = dhcp_opt_get_host_data(option);
            break;
        }
    }

    if(passed_ip == NULL) {
        FATAL_MESSAGE("No IP address passed. Cannot perform system configuration.");
    }

    if(memcmp(&interface_ip,  passed_ip, IP_ADDR_LEN)) {

        INFO_MESSAGE("using ARP to detect address collision");
        if(!arp_discover_hardware_address(dc->rawnet, 1, 5, *passed_ip,
                                          &eth_addr)) {
            ERROR_MESSAGE("DHCP server assigned us a used address. Declining.");
            client_decline(dc);

            /* sleep for up to ten seconds: fixme: make this configurable. */
            INFO_MESSAGE("sleeping before retry (for %"PRIu16" seconds)", 10);

            client_cache_delete_cache(dc->cache);
            sleep_random(10);

            return STATE_INIT;
        }

    }

    /* since we're going to use this address, let's reply for it
     * and let everyone know we're using it. */

    if(client_broadcast_arp_reply(dc, *passed_ip) < 0) {
        return STATE_FATAL_ERROR;
    }

    client_setup_timers(dc);
    do_sysconf(options, dc, STATE_SETUP);

    dhcp_opt_destroy_option_list(options);

    return STATE_BOUND;
}

/* utility state: call this if reinitializing our client control
 * useful when recreating the control (e.g. after a fork). */
int client_reinitialize(dhcp_client_control_t *dc)
{
    client_setup_timers(dc);
    return 0;
}

/******************
 * check routines *
 ******************/

/* check the requested options */
static int client_check_requested_options(dhcp_client_control_t *dc)
{
    eth_addr_t eth_addr;
    ip_addr_t ip_addr;
    char *eth_addr_string;
    char *ip_addr_string;
    uint8_t *required_options;
    uint8_t received_options[MAX_OPTIONS_HANDLED_NUM];
    dhcp_opt_t *dhcp_option;
    int i;

    /* get the mac address, and ip address from the packet. */
    ip_addr = ip_get_src_addr(dc->rawnet->ip_p);
    eth_addr = eth_get_src_addr(dc->rawnet->ether_p);

    /* find out if we have a list of required options. */
    required_options = client_conf_get_opt_required_bit_array(dc->conf, ip_addr, eth_addr);

    if(required_options == NULL) {
        WARN_MESSAGE("no required options set. accepting first response.");
        return 1;
    } 

    /* build our received options array. */
    memset(received_options, 0, sizeof(received_options));
    dhcp_reset_option_seek(dc->rawnet->dhcp_p);

    while((dhcp_option = dhcp_get_next_option(dc->rawnet->dhcp_p)) != NULL) {
        
        /* if it's in our range fill it in or ignore it and give a warning. */
        if(dhcp_opt_get_tag(dhcp_option) > MAX_OPTIONS_HANDLED) {
            WARN_MESSAGE("ignoring received option tag which is out of our range: %d", dhcp_opt_get_tag(dhcp_option));
            continue;
        }

        received_options[dhcp_opt_get_tag(dhcp_option)] = 1;
    }

    /* make sure everything in our required option list is in our requested option list. */
    for(i = 0;i < MAX_OPTIONS_HANDLED;i++) {

        if(required_options[i] && !received_options[i]) {

            /* make the strings of the addresses. */
            eth_addr_string = eth_addr_to_string(eth_addr);
            ip_addr_string = ip_addr_to_string(ip_addr);

            WARN_MESSAGE("ignoring offer from server (%s : %s) : failed to supply required option: %s", 
                         ip_addr_string, eth_addr_string, dhcp_option_printable_string_get(i));

            xfree(ip_addr_string);
            xfree(eth_addr_string);

            return 0;
        }
    }

    return 1;
}

/* basic DHCP test on incoming rawnet packet. */
static int client_check_dhcp(dhcp_client_control_t *dc)
{
    /* check for generic dhcp response.  
     * we check for type,
     * validity, matching xid, and 
     * matching server source port  */

    /* is rawnet happy with the packet? */
    if(((!rawnet_is_valid(dc->rawnet)) &&

        /* is it a DHCP packet? */
        (dc->rawnet->type != RAWNET_DHCP) &&

        /* does the XID match up? */
        (dhcp_get_xid(dc->rawnet->dhcp_p) != dc->xid) &&

        /* is it from a server? */
        (udp_get_src_port(dc->rawnet->udp_p) != dc->server_port)))
        return 0;

    /* great: valid dhcp response. */
    return 1;

}

/* check for valid discover responses. */
static int client_check_discover(void *arg)
{
    dhcp_client_control_t *dc = arg;

    /* check for a valid dhcp packet. */

    if(client_check_dhcp(dc) &&

       /* check that the type is an offer. */
       dhcp_is_type(dc->rawnet->dhcp_p, DHCP_OFFER_TM) &&

       /* check if the requested options match what we want. */
       client_check_requested_options(dc)) {
        return 1;
    }  else  {
        return 0;
    }
}

/* check for valid request responses. */
static int client_check_request(void *arg)
{
    dhcp_client_control_t *dc = arg;

    if(client_check_dhcp(dc) &&
       (dhcp_is_type(dc->rawnet->dhcp_p, DHCP_DHCPACK_TM) ||
        dhcp_is_type(dc->rawnet->dhcp_p, DHCP_DHCPNAK_TM)))
        return 1;
    else
        return 0;
}

/* check for valid renew response. */
static int client_check_renew(void *arg)
{
    /* we're expecting the same as request :-) */
    return (client_check_request(arg));

}

/* check for valid renew response. */
static int client_check_rebind(void *arg)
{
    /* we're expecting the same as request :-) */
    return (client_check_request(arg));

}

/***************************
 * client states           *
 ***************************/

/* we have an offered lease we want and are selecting it. */
int client_select(dhcp_client_control_t *dc)
{   
    list_t *dump_options;
    ip_addr_t yiaddr, siaddr;

    INFO_MESSAGE("performing DHCP SELECT");

    /* This is not relay agent friendly: FIXME later. */

    /* get the address on the packet as the server ip and hardware address. */
    dhcp_client_set_server_ip_address(dc, ip_get_src_addr(dc->rawnet->ip_p));
    dhcp_client_set_server_hw_address(dc, eth_get_src_addr(dc->rawnet->ether_p));

    /* get the yiaddr/siaddr fields. */
    yiaddr = dhcp_get_yiaddr(dc->rawnet->dhcp_p);
    siaddr = dhcp_get_siaddr(dc->rawnet->dhcp_p);

    /* if these options haven't been set, we can set them ourselves from the yiaddr/siaddr. */
    if(!dhcp_have_option(dc->rawnet->dhcp_p, TAG_DHCP_REQUESTED_IP_ADDRESS)) {
        
        list_add(dc->rawnet->dhcp_p->options,
                 dhcp_opt_create_from_internal_data(TAG_DHCP_REQUESTED_IP_ADDRESS,
                                                    &yiaddr, 
                                                    IP_ADDR_LEN));
    }

    if(!dhcp_have_option(dc->rawnet->dhcp_p, TAG_DHCP_SERVER_IDENTIFIER)) {

        list_add(dc->rawnet->dhcp_p->options, 
                 dhcp_opt_create_from_internal_data(TAG_DHCP_SERVER_IDENTIFIER,
                                                    &siaddr,
                                                    IP_ADDR_LEN));
    }

    /* before dumping options strip options we never want to dump */
    dump_options = client_remove_protocol_only_options(dc->rawnet->dhcp_p->options);

    /* dump cache file. */
    if(client_cache_dump_options(dc->cache, dump_options)) {
        ERROR_MESSAGE("could not create temporary cache.");
        return STATE_FATAL_ERROR;
    }

    dhcp_opt_destroy_option_list(dump_options);

    return STATE_REQUEST;
}

/* init state: no lease exists. */
int client_init(dhcp_client_control_t *dc)
{
    list_t *options;
    int retval;
    int retries, timeout;
    uint8_t received_response = 0;

    INFO_MESSAGE("attempting DHCP DISCOVER");

    /* get the number of retries we will attempt. */
    retries = client_conf_get_dhcp_discovery_retries(dc->conf);
    timeout = client_conf_get_dhcp_discover_timeout(dc->conf);

    /* build discovery option list. */
    options = client_build_discover_option_list(dc);

    /* when we begin transmitting discover we setup our secs
     * field.  this is because relay agents and dhcp servers can
     * use this to see how bad off we are. */
    dhcp_client_reset_secs(dc);

    /* update new xid per start of discovery . */
    dhcp_client_update_xid(dc);

    /* build our dhcp discover packet. */
    build_dhcp_discover(dc->rawnet, dc->xid, dc->secs, options);

    while(retries--) {
        retval = rawnet_packet_transact(dc->rawnet, dc, client_update_packet,
                                        client_check_discover, timeout);
        switch (retval) {

        case RAWNET_TIMEOUT:
            break;

        case RAWNET_OK:
            received_response = 1;
            break;

        case RAWNET_ERROR:
            ERROR_MESSAGE("received error from raw network handler.");
            return STATE_FATAL_ERROR;

        case RAWNET_USER_INTERRUPT:
            ERROR_MESSAGE("caught user interrupt.");
            return STATE_USER_INTERRUPT;

        default:
            FATAL_MESSAGE("invalid return value from raw network handler -- this a bug report it.");
        }

        if(received_response) /* if we got a response that passed our filters (and is good enough)
                                 then break here. */
            break;
    }

    if(received_response) {

        /* just call client_select to select the lease and store the information. */
        return client_select(dc);
    } else {

        /* otherwise we're out of retries and have not received a lease. */
        return STATE_NO_LEASES;
    }
}

static int client_requesting_proc(dhcp_client_control_t *dc, int state)
{
    list_t *options, *cache_options;
    int retval;
    int retries, timeout;
    uint8_t received_response = 0;

    INFO_MESSAGE("attempting DHCP REQUEST");

    /* get number of retries we should attempt with DHCP REQUEST. */

    retries = client_conf_get_dhcp_request_retries(dc->conf);
    timeout = client_conf_get_dhcp_request_timeout(dc->conf);

    /* Build options along with the options in our cache. */

    options = client_build_request_option_list(dc);

    if(state == STATE_INIT_REBOOT)
        cache_options = client_cache_load_options(dc->cache, 0); /* get cache. */
    else
        cache_options = client_cache_load_options(dc->cache, 1); /* get temporary cache. */

    list_join(options, cache_options); /* destroys cache_options. */
    build_dhcp_request_broadcast(dc->rawnet, dc->xid, dc->secs, 0, 0, options);

    while(retries--) {

        retval =
            rawnet_packet_transact(dc->rawnet, dc, NULL, client_check_request,
                                   timeout);

        switch (retval) {

        case RAWNET_TIMEOUT:
            break; /* timeout, do a retry. */

        case RAWNET_OK:
            received_response = 1;
            break;

        case RAWNET_ERROR:
            ERROR_MESSAGE("received error from raw network handler.");
            return STATE_FATAL_ERROR;

        case RAWNET_USER_INTERRUPT:
            ERROR_MESSAGE("caught user interrupt.");
            return STATE_USER_INTERRUPT;

        default:
            FATAL_MESSAGE("invalid return value from raw network handler -- this a bug report it.");
        }

        if(received_response)
            break;
    }

    /* check for a valid response. */
    if(received_response) {

        /* if we got a pack we're good to go, if we haven't
         * then we need to revert to discover state after waiting at least ten seconds. */

        if(dhcp_is_type(dc->rawnet->dhcp_p, DHCP_DHCPACK_TM)) {

            /* if init-reboot then record the server which replied. */
            if(state == STATE_INIT_REBOOT) {

                dhcp_client_set_server_ip_address(dc, ip_get_src_addr(dc->rawnet->ip_p));
                dhcp_client_set_server_hw_address(dc, eth_get_src_addr(dc->rawnet->ether_p));

            } else
                client_cache_update(dc->cache);

            return STATE_SETUP;

        } else {

            /* anything is else NACK because of our earlier test on the packet. */
            INFO_MESSAGE("received NACK response to request. waiting for retry.");
            client_cache_delete_tmp_cache(dc->cache);

            /* sleep for up to ten seconds: fixme: make this configurable. */
            INFO_MESSAGE("sleeping before retry (for %"PRIu16" seconds)", 10);
            sleep_random(10);

            INFO_MESSAGE("retrying.");
            return STATE_INIT;
        }

    } else {

        /* we timed out. retry immediately. */

        INFO_MESSAGE("timed out on DHCP REQUEST. attempting DISCOVER again.");
        client_cache_delete_tmp_cache(dc->cache);

        return STATE_INIT;
    }
}

/* init-reboot state: lease exists, try to acquire it. */
int client_init_reboot(dhcp_client_control_t *dc)
{
    return client_requesting_proc(dc, STATE_INIT_REBOOT);
}

/* request: request a lease. */
int client_requesting(dhcp_client_control_t *dc)
{
    return client_requesting_proc(dc, STATE_INIT);
}

/* inform: we've assigned ourselves an IP address, try to inform a dhcp server. */
int client_inform(dhcp_client_control_t *dc)
{
    /* not implemented. */
    return 0;
}

/* rebind: try to reacquire the lease. */
int client_rebinding(dhcp_client_control_t *dc)
{
    list_t *options;
    int retval;
    ip_addr_t my_addr;

    INFO_MESSAGE("attempting DHCP REBIND");

    my_addr = rawnet_get_ip_addr(dc->rawnet);

    options = client_build_rebind_option_list(dc);
    build_dhcp_request_broadcast(dc->rawnet, dc->xid, dc->secs, my_addr, 0, options);

    retval =
        rawnet_packet_transact(dc->rawnet, dc, NULL, client_check_rebind,
                               compensate_timeout_with_timer(dc));

    switch (retval) {

    case RAWNET_TIMEOUT:
        return STATE_BOUND; /* no response. we've timed out. */

    case RAWNET_OK:
        break;

    case RAWNET_ERROR:
        ERROR_MESSAGE("received error from raw network handler.");
        return STATE_FATAL_ERROR;

    case RAWNET_USER_INTERRUPT:
        ERROR_MESSAGE("caught user interrupt.");
        return STATE_USER_INTERRUPT;

    default:
        FATAL_MESSAGE("invalid return value from raw network handler -- this a bug report it.");
    }

    if(dhcp_is_type(dc->rawnet->dhcp_p, DHCP_DHCPACK_TM)) {

        /* our lease is OK. reconfigure timers. */
        client_setup_timers(dc);
        return STATE_BOUND;

    } else if(dhcp_is_type(dc->rawnet->dhcp_p, DHCP_DHCPNAK_TM)) {

        /* we've been told our lease is no good. */

        /* unconfigure everything including our interface. */
        client_unconfigure(dc);

        /* delete our cache. */
        client_cache_delete_cache(dc->cache);

        return STATE_INIT;

    } else {

        ERROR_MESSAGE("received neither NACK nor ACK -- this should never happen because of higher filters.");
        FATAL_MESSAGE("I shouldn't be here. this is a bug report me.");
        exit(1); /* get rid of compiler warning. */
    }
}

/* renew: try to renew the lease. */
int client_renewing(dhcp_client_control_t *dc)
{    
    list_t *options;
    int retval;
    ip_addr_t my_addr;

    INFO_MESSAGE("attempting DHCP RENEW");

    my_addr = rawnet_get_ip_addr(dc->rawnet);
    
    options = client_build_renew_option_list(dc);
    build_dhcp_request_unicast(dc->rawnet, dc->xid, dc->secs, options, my_addr,
                               dhcp_client_get_server_ip_address(dc),
                               0,
                               dhcp_client_get_server_hw_address(dc));

    /* find out how much time we have until the next timer.
     * if we have no other timers it's an error, however
     * it will be caught later now just do our best. */

    retval =
        rawnet_packet_transact(dc->rawnet, dc, NULL, client_check_renew,
                               compensate_timeout_with_timer(dc));

    switch (retval) {

    case RAWNET_TIMEOUT:
        return STATE_BOUND; /* no response: we've timed out. */

    case RAWNET_OK:
        break;

    case RAWNET_ERROR:
        ERROR_MESSAGE("received error from raw network handler.");
        return STATE_FATAL_ERROR;

    case RAWNET_USER_INTERRUPT:
        ERROR_MESSAGE("caught user interrupt.");
        return STATE_USER_INTERRUPT;

    default:
        FATAL_MESSAGE
            ("dhcp-client-states: invalid return value from raw network handler -- this a bug report it.");
    }

    if(dhcp_is_type(dc->rawnet->dhcp_p, DHCP_DHCPACK_TM)) {

        /* our lease is OK. reconfigure timers. */
        client_setup_timers(dc);
        return STATE_BOUND;

    } else if(dhcp_is_type(dc->rawnet->dhcp_p, DHCP_DHCPNAK_TM)) {

        /* we've been told our lease is no good. */

        /* unconfigure everything including our interface. */
        client_unconfigure(dc);

        /* delete our cache. */
        client_cache_delete_cache(dc->cache);

        return STATE_INIT;

    } else {

        ERROR_MESSAGE("received neither NACK nor ACK -- this should never happen because of higher filters.");
        FATAL_MESSAGE("I shouldn't be here. this is a bug report me.");
        exit(1); /* get rid of compiler warning. */
    }
}

/* we're bound and should wait for lease expiry or timers. */
int client_bound(dhcp_client_control_t *dc)
{
    uint32_t next_timer = 0; /* get rid of compiler warnings. */
    time_t time_before = 0; /* get rid of compiler warnings. */
    time_t time_after, time_interval;

    INFO_MESSAGE("entering WAIT stage");

    /* take our raw network connection down. */
    rawnet_down(dc->rawnet);

    if(dc->lease_time_is_infinite) {

        INFO_MESSAGE("lease time is infinite. waiting forever.");

        /* wait for any interrupts. */
        suspend_for_interrupts();

    } else {

        /* peek and store our next timer in case we get an interrupt
         * which is not a timer interrupt. */

        next_timer = timer_peek_next_timer(dc->context->timer);
        time_before = time(NULL);

        /* now fire off the next timer. */
        if(timer_set_next(dc->context->timer)) {
            FATAL_MESSAGE("no timers set. we should have at least one for lease expiry sent to us from the server.");
        }

        /* wait for any interrupts: we're OK. there is no race
         * condition here since our suspension will unmask us
         * and allow any previous signals to affect us here. */

        suspend_for_interrupts();

    }
    
    /* at this point we may or may not have received a timer
     * interrupt. if we have not then setup a timer to go off
     * once we return by storing a new timer with the time
     * difference. */

    if(peek_interrupt_type() != INTERRUPT_ALARM && !dc->lease_time_is_infinite) {

        time_after = time(NULL);
        time_interval = time_after - time_before; /* number of seconds elapsed. */

        if(time_interval < 0) {

            /* someone has been b0rking the system clock. */
            WARN_MESSAGE("the system clock is skewed or you've been messing with it. i'm going to just use the next timer and hope for the best.");

        } else {

            /* we did not receive a timer interrupt.
             * put the difference as the next timer to go off. */

            next_timer -= time_interval; /* this is safe since we're promoting the type
                                          * and it can never be < 0 */

            if(next_timer == 0)
                next_timer = 1; /* give at least one second. */

            /* we just set our next timer to atleast one. this
             * has one implication. if we're firing off other
             * signals to the daemon (HUP for example) all the
             * time it will never get the alarm.  opinion: you
             * shouldn't HUP your daemon constantly.  otherwise
             * we at most lag by one second which is not bad. */

            timer_add_trigger(dc->context->timer, next_timer, timer_get_current_id(dc->context->timer));
        }
    }

    /* bring the raw network devices back up.  we'll probably
     * need them: FIXME, revamp into one network module that uses
     * raw or user level networking depending on our interface's
     * state. */

    rawnet_up(dc->rawnet);

    return STATE_BOUND; /* we want to return back to this state if the interrupt wasn't an alarm. */
}

/* release a lease. */
int client_release(dhcp_client_control_t *dc)
{
    list_t *cache_options, *options;
    ip_addr_t *my_addr = NULL;
    ip_addr_t ciaddr;
    dhcp_opt_t *option;

    INFO_MESSAGE("sending DHCP RELEASE");

    cache_options = client_cache_load_options(dc->cache, 0);

    list_rewind(cache_options);
    while((option = list_next(cache_options)) != NULL) {

        if(dhcp_opt_get_tag(option) == TAG_DHCP_REQUESTED_IP_ADDRESS) {
            my_addr = dhcp_opt_get_host_data(option);
            break;
        }
    }

    if(my_addr == NULL)  { /* if empty, don't release. */
        dhcp_opt_destroy_option_list(cache_options);
        return STATE_FATAL_ERROR;
    }

    ciaddr = *my_addr;
    dhcp_opt_destroy_option_list(cache_options);

    options = client_build_release_option_list(dc);

    build_dhcp_release(dc->rawnet, dc->xid, options, ciaddr,
                       dhcp_client_get_server_ip_address(dc),
                       dhcp_client_get_server_hw_address(dc));

    if(rawnet_send_packet(dc->rawnet) < 0) {
        ERROR_MESSAGE("could not send RELEASE.");
        return STATE_FATAL_ERROR;
    }

    /* called from do_shutdown() so no real state returned.
     * however if we do take the return value into consideration
     * we should handle it as an error. */
    return STATE_FATAL_ERROR;
}

/* perform system configuration. */
int client_setup(dhcp_client_control_t *dc)
{
    INFO_MESSAGE("entering SETUP stage");
    return client_configure(dc);
}

/* do a shutdown. */
int client_shutdown(dhcp_client_control_t *dc)
{
    INFO_MESSAGE("shutting down.");

    if(dc) { /* only do clean up if we can. */

        /* send our a release only if we have a cache. */
        if(!client_cache_is_empty(dc->cache))
            client_release(dc);

        /* unconfigure everything including our interface. */
        client_unconfigure(dc);

        /* since we're releasing the lease we should down the interface */

        INFO_MESSAGE("bringing down interface.");
        rawnet_down(dc->rawnet);
        rawnet_interface_down(dc->rawnet);

        /* delete the PID file. */
        file_delete_pid(dc->interface);

        /* delete our cache. */
        client_cache_delete_cache(dc->cache);
        client_cache_delete_tmp_cache(dc->cache);

        dhcp_client_control_destroy(dc);
    }

    exit(0);
}

/* decline a lease. */
int client_decline(dhcp_client_control_t *dc)
{
    list_t *options, *decline_options;
    dhcp_opt_t *option;
    ip_addr_t *passed_ip = NULL;

    INFO_MESSAGE("performing DHCP decline");

    options = client_cache_load_options(dc->cache, 0);

    list_rewind(options);
    while((option = list_next(options)) != NULL) {

        if(dhcp_opt_get_tag(option) == TAG_DHCP_REQUESTED_IP_ADDRESS) {
            passed_ip = dhcp_opt_get_host_data(option);
            break;
        }
    }

    if(passed_ip == NULL) {
        FATAL_MESSAGE("no address assigned and we're trying to decline. this is a bug, report me.");
    }

    decline_options = client_build_decline_option_list(dc, *passed_ip);
    dhcp_opt_destroy_option_list(options);

    build_dhcp_decline(dc->rawnet, dc->xid, dc->secs, decline_options);
    if(rawnet_send_packet(dc->rawnet) < 0) {
        ERROR_MESSAGE("could not send DECLINE.");
        return STATE_FATAL_ERROR;
    }

    /* a lot like release, our return here shouldn't matter since
     * we are most likely called within a state. if not then the
     * state simply shouldn't be handled. */

    return STATE_FATAL_ERROR;
}
