/* $Header: /cvsroot/dhcp-agent/dhcp-agent/src/dhcp-client.c,v 1.35 2003/07/06 17:14:43 actmodern Exp $
 * 
 * Copyright 2002 Thamer Alharbash <tmh@whitefang.com>
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in
 * the documentation and/or other materials provided with the
 * distribution.
 * 3. The names of the authors may not be used to endorse or promote
 * products derived from this software without specific prior
 * written permission.
 * 
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 * 
 */

#define MODULE_NAME "dhcp-client"

#include "dhcp-local.h"
#include "dhcp-limits.h"
#include "dhcp-libutil.h"
#include "dhcp-librawnet.h"

#include "dhcp-client-conf.h"
#include "dhcp-cache-entry.h"
#include "dhcp-client-cache.h"
#include "dhcp-client.h"
#include "dhcp-client-guile.h"
#include "dhcp-daemon.h"

/* forward declarations of do_ routines. */
static void do_version(char *interface);
static void do_kill_client(char *interface);
static void do_wake_client(char *interface);
static void do_clear_cache(char *interface);
static void do_client(char *interface);
static void do_status(char *interface);

/* command to string table. */
const char *command_string[] = {
    "version", "kill",
    "wake", "clear", "do client"
};

/* table of functions indexed by command codes. */
client_command commands[] = {
    do_version,                 /* print out version information. */
    do_kill_client,             /* kill client. */
    do_wake_client,             /* wake client. */
    do_clear_cache,             /* clear cache. */
    do_client,                  /* initialize and run client. */
    do_status,                  /* get status info. */
};

/* forward declaration for interface routines. */
static char *interface_get_active(void);
static char *interface_get_inactive(void);

/* Our get interface sister routine.
   Different command codes will want
   to select interface on different
   criteria. We use this to manage it. */

interface_lister get_interface[] = {
    NULL,                       /* version output needs nothing. */
    interface_get_active,       /* kill client. */
    interface_get_active,       /* wake client. */
    interface_get_active,       /* clear cache. */
    interface_get_inactive,     /* run client. */
    interface_get_active,       /* get status. */
};

/* client state dispatch: index constants in dhcp-client.h */
client_state client_states[] = {
    client_init,
    client_init_reboot,
    client_requesting,
    client_renewing,
    client_rebinding,
    client_setup,
    client_bound,
    client_inform,
    client_shutdown,
    client_release,
    client_decline,
};

/* global vars affecting other code. */
char *work_dir = DHCPLOCALSTATE_CLIENTDIR; /* our default working directory */

#if !defined(HAVE_PROGNAME)
const char *__progname;
#endif                          /* HAVE_PROGNAME */

/* local vars for dhcp-client. */
static char *interface = NULL;  /* the interface we're concerned with. */
static unsigned char want_background = 1; /* whether we should go into the background. */
static char *fake_hw_addr = NULL; /* if using fake hardware address set here. */
static int promiscuous = 0;     /* whether or not we're running in promiscious mode. */

/* Utility routines. */

/* initialize state per contents of cache. */
static int initialize_client_state(dhcp_client_control_t *dc)
{
    if(client_cache_is_empty(dc->cache)) {
        return STATE_INIT;
    } else /* if not empty we assume we can request the options. */
        return STATE_INIT_REBOOT;
}

/* print usage info. */
static void usage(char *s)
{
    printf("usage: %s [-scpavkwt] [-l verbosity level] [-d directory]\n", s);
    printf("usage: %-*s [-i interface name] [ -m mac address ]\n", strlen(s), " ");
    exit(0);
}

/* print out version information and exit. */
static void do_version(char *interface)
{
    INFO_MESSAGE("%s version: %s\n", getprogname(), VERSION);
}

/* check for existing client. */
static int client_process_exists(char *interface)
{
    pid_t pid;

    if(file_get_pid(interface, &pid))
        return 0; /* no pid file -- therefore no process. */

#if !defined(KILL_SIGNAL_DETECT)

    /* if we can't detect it assume it is up since the PID file exists :| 
     * issue warning though. */
    WARN_MESSAGE("%s PID file exists please delete and kill process.", getprogname());
    return 0;

#else
    if(!kill(pid, 0))
        return 1;
    else {
        file_delete_pid(interface);
        return 0;
    }
#endif

}

/* wrapper around interface retrieval routines. */

/* get active or inactive interfaces. */
static char *interface_get_proc(int active)
{
    char *first_interface;
    list_t *interface_list;

    if(active)
        interface_list = rawnet_list_active_interfaces();
    else
        interface_list = rawnet_list_inactive_interfaces();

    if(list_get_len(interface_list) == 0) {
        list_destroy(interface_list, xfree);
        return NULL;
    }

    first_interface = xstrdup(list_first(interface_list));
    list_destroy(interface_list, xfree);
    return first_interface;
}

/* get interface which is active. */
static char *interface_get_active(void)
{
    return (interface_get_proc(1));
}

/* get interface which is inactive. */
static char *interface_get_inactive(void)
{
    return (interface_get_proc(0));
}

/* main loop: if only_setup is called we exit after setup --
 * this is useful for running through the loop only long
 * enough to setup the interface. */
static int do_client_dhcp_loop(int only_setup, dhcp_client_control_t *dc, int state)
{

    while(1) {

        /* In between states we're clean.
         * 
         * If we get interrupted inside a state
         * we still should be clean, but we don't
         * want to shutdown if we can help it.
         * 
         * Handle any interrupts here. */

        while(check_for_interrupts()) {


            switch (got_interrupt_type()) {

            case INTERRUPT_NONE:
                /* no more interrupts. we're done. */
                break;

            case INTERRUPT_TERM:
                state = STATE_SHUTDOWN;
                break;

            case INTERRUPT_HUP:
                /* re-read configuration. */

                INFO_MESSAGE("re-reading configuration");
                if(client_conf_reread(dc->conf)) {
                    WARN_MESSAGE("could not read configuration. using previous configuration.");
                }

                break;

            case INTERRUPT_ALARM:

                /* if alarm then check for timer type so we can set our state accordingly. */

                switch (timer_get_current_id(dc->context->timer)) {

                case TIMER_RENEW:
                    state = STATE_RENEW;
                    break;

                case TIMER_IP_LEASE:
                    state = STATE_INIT;
                    break;

                case TIMER_REBIND:
                    state = STATE_REBIND;
                    break;

                default:
                    FATAL_MESSAGE("invalid timer type caught. this is a bug. report me.");
                }
                break;

            default:
                FATAL_MESSAGE("invalid interrupt type caught. this is a bug. report me.");
            }
        }

        /* if we're only running to be setup, 
         * return here. */

        if((state == STATE_BOUND) && only_setup) {
            return state;
        }

        if(state == STATE_USER_INTERRUPT || state == STATE_NO_LEASES || state == STATE_FATAL_ERROR)
            state = STATE_SHUTDOWN;

        state = client_states[state] (dc);

    }

    return STATE_FATAL_ERROR; /* we should never get here, but if we do it's bad :) */
}

/* kill a running client. */
static void do_kill_client(char *interface)
{
    pid_t pid;

    INFO_MESSAGE("killing %s running on interface: %s", getprogname(), interface);

    if(file_get_pid(interface, &pid)) {
        ERROR_MESSAGE("could not get PID for client %s. maybe it's not running?", interface);
        return;
    }

    if(kill(pid, SIGTERM) < 0) {
        ERROR_MESSAGE("coud not send signal to client on interface: %s. maybe it's not running?",
                      interface);
        return;
    }

    INFO_MESSAGE("killed client on pid: %u", pid);
    return;
}

/* wake up a client. send sigalarm. */
static void do_wake_client(char *interface)
{
    pid_t pid;

    if(file_get_pid(interface, &pid)) {
        ERROR_MESSAGE
            ("could not get PID for client running on interface %s. Maybe it's not running?", interface);
        return;
    }

    if(kill(pid, SIGALRM) < 0) {
        ERROR_MESSAGE("could not send signal to client on %s. Maybe it's not running?", interface);
        return;
    }

    INFO_MESSAGE("woke client on pid %u : for interface: %s", pid, interface);
    return;
}

/* delete existing cache. make sure no other client is running. */
static void do_clear_cache(char *interface)
{
    pid_t pid;
    client_cache_t *cc;

    if(!file_get_pid(interface, &pid)) {
        INFO_MESSAGE("%s already is running on %s. I won't delete the cache until it shuts down.",
                     getprogname(), interface);
        return;
    }

    cc = client_cache_create(interface);

    /* now delete. */
    client_cache_delete_cache(cc);
    client_cache_delete_tmp_cache(cc);

    return;
}

static void do_status(char *interface)
{
    /* try to load up cache. */
    client_cache_t *cache;

    if((cache = client_cache_create(interface)) == NULL) {
        FATAL_MESSAGE("could not create client cache");
    }

    if(client_cache_is_empty(cache)) {

        if(client_cache_is_empty_tmp(cache)) {


            INFO_MESSAGE("No lease acquired for this interface.");
            INFO_MESSAGE(" ");

        } else {

            INFO_MESSAGE("This lease has been offered but has not been acquired yet, or the client has not been able to receive an acknowledgement from the server yet.");
            INFO_MESSAGE(" ");
            client_pretty_print_cache(cache, 1);

        }

    } else {

        client_pretty_print_cache(cache, 0);
    }

    return;
}

/* do client: perform dhcp client proper functions. */
static void do_client(char *interface)
{
    dhcp_client_control_t *dc;
    dhcp_client_control_context_t *dc_context;
    int state;

    INFO_MESSAGE("starting for interface %s", interface);

    /* Now see if another client is running on the same interface. */

    if(client_process_exists(interface)) {
        FATAL_MESSAGE("%s already exists on interface %s use -k to kill it", getprogname(), interface);
    }

    /* setup umask: this is also done in go_background but the client
     * doesn't go into the background immediately. so do it now. */

    umask(0);

    /* 
     * Create our control object.
     * There are several things it does
     * for us:
     * 
     * Read the cache, and initialize our state.
     * Read our configuration.
     * Acquire a handle on the interface control.
     * Acquire a handle on the rawnet driver.
     * 
     */

    /* "and she came from a great Desktop; and she razed the
     *  posers leaving a blanket of ash; and we were Praised in
     *  our Judgements;" -- from the Book of Ada (blessed be her
     *  semicolons)  */

    setup_interrupt_handlers(); /* setup signal handling */

    if((dc = dhcp_client_control_create(interface, promiscuous, 1)) == NULL) {
        ERROR_MESSAGE("encountered a fatal error. I'm exiting.");
        client_states[STATE_SHUTDOWN] (dc);
    }

    /* load up user guile sysconf script. */
    dhcp_guile_init(dc);

    if(fake_hw_addr) /* Marla, you liar, you big tourist, I need this. Now get out! */
        dhcp_client_control_use_fake_hw_addr(dc, fake_hw_addr);

    /* We're good to go. Look up state. */
    state = initialize_client_state(dc);

    /* do first loop, returning after setup. */
    state = do_client_dhcp_loop(1, dc, state);

    /* flush stdout to get messages out _now_
     * in case things break in forking into child. */
    fflush(stdout);
    fflush(stderr);

    /* Go in background now before main loop. */
    if(want_background) { /* are we interactive? if so stay in foreground. 
                           * ... otherwise ... */

        /* We need to destroy our client control object before
         * backgrounding ourselves.  Since we close file
         * descriptors and enter a different process space.
         *
         * This accumilates to losing a _lot_ of things:
         *
         * (1) Any raw network devices, or anything else which
         *     has a descriptor open is now closed.
         *
         * (2) Our signal handlers are gone.
         *
         * (3) Our concept of timers is gone. 
         *
         * The only thing we keep is our state which ought to be
         * STATE_BOUND
         *
         * Unfortunately destroying the control means valuable
         * data like timers, and addresses are lost.  we can
         * however make a copy of this information and just label
         * it a context. This way we set it after recreating the
         * control
         */

        dc_context = dhcp_client_control_copy_context(dc);
        dhcp_client_control_destroy(dc);

        set_interactive(0);
        go_background(work_dir);
        setup_interrupt_handlers(); /* setup signal handling,
                                     * we lose them after we fork in go_background. */

        /* set clear_interface to 0 on create call because by now we're configured fine. */
        if((dc = dhcp_client_control_create(interface, promiscuous, 0)) == NULL) {
            ERROR_MESSAGE("encountered a fatal error. I'm exiting.", interface);
            client_states[STATE_SHUTDOWN] (NULL);
        }

        /* set old context information. */
        dhcp_client_control_set_context(dc, dc_context);

        if(file_create_pid(interface)) {
            ERROR_MESSAGE("could not create PID file for interface: %s", interface);
            client_states[STATE_SHUTDOWN](NULL);
        }

        if(fake_hw_addr) /* Marla, you liar, you big tourist, I need this. Now get out! */
            dhcp_client_control_use_fake_hw_addr(dc, fake_hw_addr);

        /* call reinitialize for things like resetting the timer. */
        client_reinitialize(dc);
    }

    do_client_dhcp_loop(0, dc, state); /* perform infinite waiting loop. */
    exit(0);
}

/* main function: called from scm_boot_guile */
static void real_main(void *closure, int argc, char *argv[])
{
    int c, command_code = DO_CLIENT;

#if !defined(HAVE_PROGNAME)
    __progname = argv[0];
#endif                          /* HAVE_PROGNAME */

    INFO_MESSAGE("(C) 2003 Thamer Al-Harbash <tmh@whitefang.com>");
    INFO_MESSAGE("See LICENSE file for details.");
    INFO_MESSAGE(" ");

    while((c = getopt(argc, argv, "stpcdavi:m:kwhl:")) != -1) {

        switch (c) {

            /* check for command codes. */
        case 'v': /* print version. */
            command_code = DO_VERSION;
            break;

        case 'k': /* kill client. */
            command_code = DO_KILL;
            break;

        case 'w': /* wake client. */
            command_code = DO_WAKE;
            break;

        case 'c': /* clear up cache. */
            command_code = DO_CLEAR;
            break;

            /* setup options. */

        case 'a':
            want_background = 0;
            break;

        case 'i':
            interface = xstrdup(optarg);
            break;

        case 'm':
            fake_hw_addr = xstrdup(optarg);
            break;

        case 'l':

            if(!is_signed_numeric(optarg) || set_verbosity_level(atoi(optarg))) {
                ERROR_MESSAGE("illegal verbosity level: %s", optarg);
                exit(1);
            }

            break;

        case 's':
            command_code = DO_STATUS;
            break;

        case 'p':
            promiscuous = 1;
            break;

        case 'h': /* fall through. */
        default:
            usage(argv[0]);
            break;
        }
    }

    /* if all the user wants is version information, give it to him. */
    if(command_code == DO_VERSION) {
        do_version(NULL);
        exit(0);
    }

    do_change_work_dir(work_dir);

    /* get interface if not specified. */
    if(interface == NULL) {
        if(get_interface[command_code] != NULL)
            interface = get_interface[command_code] ();
    }

    if(interface == NULL && command_code != DO_VERSION) {
        FATAL_MESSAGE("unable to retrieve suitable interface for this operation: %s",
                      command_string[command_code]);
    }

    commands[command_code] (interface);
    exit(0);
}

/* actual main: boot guile up so it knows where the bottom of the stack is. */
int main(int argc, char *argv[])
{
    scm_boot_guile(argc, argv, real_main, NULL);

    /* get rid of compiler warning. */
    exit(1);
}
