/* $Header: /cvsroot/dhcp-agent/dhcp-agent/src/dhcp-option.c,v 1.6 2003/06/27 03:16:27 actmodern Exp $
 * 
 * Copyright 2002, 2003 Thamer Alharbash <tmh@whitefang.com>
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in
 * the documentation and/or other materials provided with the
 * distribution.
 * 3. The names of the authors may not be used to endorse or promote
 * products derived from this software without specific prior
 * written permission.
 * 
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 * 
 */

#define MODULE_NAME "dhcp-option"

#include "dhcp-local.h"
#include "dhcp-libutil.h"
#include "dhcp-librawnet.h"

#include "dhcp-option.h"
#include "dhcp-option-convert.h"

#include "dhcp-limits.h"

/* forward declaration of option destructors. */

static void dhcp_opt_destroy_string(dhcp_opt_t *opt);
static void dhcp_opt_destroy_array(dhcp_opt_t *opt);
static void dhcp_opt_destroy_atom(dhcp_opt_t *opt);
static void dhcp_opt_destroy_list(dhcp_opt_t *opt);

dhcp_opt_attr_t dhcp_opt_attr[] = {

    /* Pad: we never actually handle this here. so sit all handlers to NULL. */
    { TAG_DHCP_PAD,
      sizeof(uint8_t),
      DHCP_OPT_ATOM,
      DHCP_OPT_VAL_UINT8,
      NULL,
      NULL,
      NULL,
      NULL, 
      NULL,
      NULL,
      dhcp_opt_destroy_atom,
    },

    /* Subnet Mask */
    { TAG_DHCP_SUBNET_MASK,
      IP_ADDR_LEN,
      DHCP_OPT_ATOM,
      DHCP_OPT_VAL_ADDRESS,
      dhcp_opt_to_network_ip_addr,
      dhcp_opt_to_user_string_ip_addr,
      dhcp_opt_to_internal_string_ip_addr,
      dhcp_opt_from_network_ip_addr,
      dhcp_opt_from_user_string_ip_addr,
      dhcp_opt_from_internal_string_ip_addr,
      dhcp_opt_destroy_atom,
    },

    /* Time Offset Option */
    { TAG_DHCP_TIME_OFFSET,
      sizeof(int32_t),
      DHCP_OPT_ATOM,
      DHCP_OPT_VAL_INT32,
      dhcp_opt_to_network_int32,
      dhcp_opt_to_user_string_int32,
      dhcp_opt_to_internal_string_int32,
      dhcp_opt_from_network_int32,
      dhcp_opt_from_user_string_int32,
      dhcp_opt_from_internal_string_int32,
      dhcp_opt_destroy_atom,
    },

    /* Router Option */
    { TAG_DHCP_ROUTER,
      IP_ADDR_LEN,
      DHCP_OPT_LIST,
      DHCP_OPT_VAL_ADDRESS,
      dhcp_opt_to_network_list_ip_addr,
      dhcp_opt_to_user_string_list_ip_addr,
      dhcp_opt_to_internal_string_list_ip_addr,
      dhcp_opt_from_network_list_ip_addr,
      dhcp_opt_from_user_string_list_ip_addr,
      dhcp_opt_from_internal_string_list_ip_addr,
      dhcp_opt_destroy_list,
    },

    /* Time Servers Option */
    { TAG_DHCP_TIME_SERVER,
      IP_ADDR_LEN,
      DHCP_OPT_LIST,
      DHCP_OPT_VAL_ADDRESS,
      dhcp_opt_to_network_list_ip_addr,
      dhcp_opt_to_user_string_list_ip_addr,
      dhcp_opt_to_internal_string_list_ip_addr,
      dhcp_opt_from_network_list_ip_addr,
      dhcp_opt_from_user_string_list_ip_addr,
      dhcp_opt_from_internal_string_list_ip_addr,
      dhcp_opt_destroy_list,
    },

    /* Name Servers Option */
    { TAG_DHCP_NAME_SERVER,
      IP_ADDR_LEN,
      DHCP_OPT_LIST,
      DHCP_OPT_VAL_ADDRESS,
      dhcp_opt_to_network_list_ip_addr,
      dhcp_opt_to_user_string_list_ip_addr,
      dhcp_opt_to_internal_string_list_ip_addr,
      dhcp_opt_from_network_list_ip_addr,
      dhcp_opt_from_user_string_list_ip_addr,
      dhcp_opt_from_internal_string_list_ip_addr,
      dhcp_opt_destroy_list,
    },

    /* Domain Name Servers Option */
    { TAG_DHCP_DOMAIN_NAME_SERVER,
      IP_ADDR_LEN,
      DHCP_OPT_LIST,
      DHCP_OPT_VAL_ADDRESS,
      dhcp_opt_to_network_list_ip_addr,
      dhcp_opt_to_user_string_list_ip_addr,
      dhcp_opt_to_internal_string_list_ip_addr,
      dhcp_opt_from_network_list_ip_addr,
      dhcp_opt_from_user_string_list_ip_addr,
      dhcp_opt_from_internal_string_list_ip_addr,
      dhcp_opt_destroy_list,
    },

    /* Log Servers Option */
    { TAG_DHCP_LOG_SERVER,
      IP_ADDR_LEN,
      DHCP_OPT_LIST,
      DHCP_OPT_VAL_ADDRESS,
      dhcp_opt_to_network_list_ip_addr,
      dhcp_opt_to_user_string_list_ip_addr,
      dhcp_opt_to_internal_string_list_ip_addr,
      dhcp_opt_from_network_list_ip_addr,
      dhcp_opt_from_user_string_list_ip_addr,
      dhcp_opt_from_internal_string_list_ip_addr,
      dhcp_opt_destroy_list,
    },

    /* Cookie Servers Option */
    { TAG_DHCP_COOKIE_SERVER,
      IP_ADDR_LEN,
      DHCP_OPT_LIST,
      DHCP_OPT_VAL_ADDRESS,
      dhcp_opt_to_network_list_ip_addr,
      dhcp_opt_to_user_string_list_ip_addr,
      dhcp_opt_to_internal_string_list_ip_addr,
      dhcp_opt_from_network_list_ip_addr,
      dhcp_opt_from_user_string_list_ip_addr,
      dhcp_opt_from_internal_string_list_ip_addr,
      dhcp_opt_destroy_list,
    },

    /* LPR Servers Option */
    { TAG_DHCP_LPR_SERVER,
      IP_ADDR_LEN,
      DHCP_OPT_LIST,
      DHCP_OPT_VAL_ADDRESS,
      dhcp_opt_to_network_list_ip_addr,
      dhcp_opt_to_user_string_list_ip_addr,
      dhcp_opt_to_internal_string_list_ip_addr,
      dhcp_opt_from_network_list_ip_addr,
      dhcp_opt_from_user_string_list_ip_addr,
      dhcp_opt_from_internal_string_list_ip_addr,
      dhcp_opt_destroy_list,
    },

    /* Impress Servers Option */
    { TAG_DHCP_IMPRESS_SERVER,
      IP_ADDR_LEN,
      DHCP_OPT_LIST,
      DHCP_OPT_VAL_ADDRESS,
      dhcp_opt_to_network_list_ip_addr,
      dhcp_opt_to_user_string_list_ip_addr,
      dhcp_opt_to_internal_string_list_ip_addr,
      dhcp_opt_from_network_list_ip_addr,
      dhcp_opt_from_user_string_list_ip_addr,
      dhcp_opt_from_internal_string_list_ip_addr,
      dhcp_opt_destroy_list,
    },

    /* Resource Location Servers Option */
    { TAG_DHCP_RESOURCE_LOCATION_SERVER,
      IP_ADDR_LEN,
      DHCP_OPT_LIST,
      DHCP_OPT_VAL_ADDRESS,
      dhcp_opt_to_network_list_ip_addr,
      dhcp_opt_to_user_string_list_ip_addr,
      dhcp_opt_to_internal_string_list_ip_addr,
      dhcp_opt_from_network_list_ip_addr,
      dhcp_opt_from_user_string_list_ip_addr,
      dhcp_opt_from_internal_string_list_ip_addr,
      dhcp_opt_destroy_list,
    },

    /* Host name Option */
    { TAG_DHCP_HOST_NAME,
      sizeof(int8_t),
      DHCP_OPT_STRING,
      DHCP_OPT_VAL_INT8,
      dhcp_opt_to_network_nvt_string,
      dhcp_opt_to_user_string_nvt_string,
      dhcp_opt_to_internal_string_nvt_string,
      dhcp_opt_from_network_nvt_string, /* FIXME: see RFC 1035 for character set restriction. */
      dhcp_opt_from_user_string_nvt_string,
      dhcp_opt_from_internal_string_nvt_string,
      dhcp_opt_destroy_string,
    },

    /* Boot File Size Option. */
    { TAG_DHCP_BOOT_FILE_SIZE,
      sizeof(uint16_t),
      DHCP_OPT_ATOM,
      DHCP_OPT_VAL_UINT16,
      dhcp_opt_to_network_int16,
      dhcp_opt_to_user_string_int16,
      dhcp_opt_to_internal_string_int16,
      dhcp_opt_from_network_int16,
      dhcp_opt_from_user_string_int16,
      dhcp_opt_from_internal_string_int16,
      dhcp_opt_destroy_atom,
    },

    /* Merit Dump File Option. */
    { TAG_DHCP_MERIT_DUMP_FILE,
      sizeof(int8_t),
      DHCP_OPT_STRING,
      DHCP_OPT_VAL_INT8,
      dhcp_opt_to_network_nvt_string,
      dhcp_opt_to_user_string_nvt_string,
      dhcp_opt_to_internal_string_nvt_string,
      dhcp_opt_from_network_nvt_string,
      dhcp_opt_from_user_string_nvt_string,
      dhcp_opt_from_internal_string_nvt_string,
      dhcp_opt_destroy_string,
    },

    /* Domain Name Option. */
    { TAG_DHCP_DOMAIN_NAME,
      sizeof(int8_t),
      DHCP_OPT_STRING,
      DHCP_OPT_VAL_INT8,
      dhcp_opt_to_network_string,
      dhcp_opt_to_user_string_string,
      dhcp_opt_to_internal_string_string,
      dhcp_opt_from_network_string,
      dhcp_opt_from_user_string_string,
      dhcp_opt_from_internal_string_string,
      dhcp_opt_destroy_string,
    },

    /* Swap Server Option */
    { TAG_DHCP_SWAP_SERVER,
      IP_ADDR_LEN,
      DHCP_OPT_ATOM,
      DHCP_OPT_VAL_ADDRESS,
      dhcp_opt_to_network_ip_addr,
      dhcp_opt_to_user_string_ip_addr,
      dhcp_opt_to_internal_string_ip_addr,
      dhcp_opt_from_network_ip_addr,
      dhcp_opt_from_user_string_ip_addr,
      dhcp_opt_from_internal_string_ip_addr,
      dhcp_opt_destroy_atom,
    },

    /* Root Path Option */
    { TAG_DHCP_ROOT_PATH,
      sizeof(int8_t),
      DHCP_OPT_STRING,
      DHCP_OPT_VAL_INT8,
      dhcp_opt_to_network_nvt_string,
      dhcp_opt_to_user_string_nvt_string,
      dhcp_opt_to_internal_string_nvt_string,
      dhcp_opt_from_network_nvt_string,
      dhcp_opt_from_user_string_nvt_string,
      dhcp_opt_from_internal_string_nvt_string,
      dhcp_opt_destroy_string,
    },

    /* Extensions Path Option */
    { TAG_DHCP_EXTENSIONS_PATH,
      sizeof(int8_t),
      DHCP_OPT_STRING,
      DHCP_OPT_VAL_INT8,
      dhcp_opt_to_network_string,
      dhcp_opt_to_user_string_string,
      dhcp_opt_to_internal_string_string,
      dhcp_opt_from_network_string,
      dhcp_opt_from_user_string_string,
      dhcp_opt_from_internal_string_string,
      dhcp_opt_destroy_string,
    },

    /* IP Forwarding Option */
    { TAG_DHCP_IP_FORWARDING,
      sizeof(uint8_t),
      DHCP_OPT_ATOM,
      DHCP_OPT_VAL_UINT8,
      dhcp_opt_to_network_uint8,
      dhcp_opt_to_user_string_uint8,
      dhcp_opt_to_internal_string_uint8,
      dhcp_opt_from_network_uint8,
      dhcp_opt_from_user_string_uint8,
      dhcp_opt_from_internal_string_uint8,
      dhcp_opt_destroy_atom,
    },

    /* Non Local Source Routing Option */
    { TAG_DHCP_NON_LOCAL_SOURCE_ROUTING,
      sizeof(uint8_t),
      DHCP_OPT_ATOM,
      DHCP_OPT_VAL_UINT8,
      dhcp_opt_to_network_uint8,
      dhcp_opt_to_user_string_uint8,
      dhcp_opt_to_internal_string_uint8,
      dhcp_opt_from_network_uint8,
      dhcp_opt_from_user_string_uint8,
      dhcp_opt_from_internal_string_uint8,
      dhcp_opt_destroy_atom,
    },

    /* Policy Filter Option */
    { TAG_DHCP_POLICY_FILTER,
      IP_ADDR_LEN,
      DHCP_OPT_LIST,
      DHCP_OPT_VAL_ADDRESS_PAIR,
      dhcp_opt_to_network_ip_addr_pair,
      dhcp_opt_to_user_string_ip_addr_pair,
      dhcp_opt_to_internal_string_ip_addr_pair,      
      dhcp_opt_from_network_list_ip_addr_pair,
      dhcp_opt_from_user_string_list_ip_addr,
      dhcp_opt_from_internal_string_list_ip_addr,
      dhcp_opt_destroy_list,
    },

    /* Maximum Datagram Reassembly Size Option */
    { TAG_DHCP_MAX_DGRAM_REASSUMBLY_SIZE,
      sizeof(uint16_t),
      DHCP_OPT_ATOM,
      DHCP_OPT_VAL_UINT16,
      dhcp_opt_to_network_uint16,
      dhcp_opt_to_user_string_uint16,
      dhcp_opt_to_internal_string_uint16,
      dhcp_opt_from_network_uint16,
      dhcp_opt_from_user_string_uint16,
      dhcp_opt_from_internal_string_uint16,
      dhcp_opt_destroy_atom,
    },

    /* Default IP Time-to-Live Option */
    { TAG_DHCP_IP_TIME_TO_LIVE,
      sizeof(uint8_t),
      DHCP_OPT_ATOM,
      DHCP_OPT_VAL_UINT8,
      dhcp_opt_to_network_uint8,
      dhcp_opt_to_user_string_uint8,
      dhcp_opt_to_internal_string_uint8,
      dhcp_opt_from_network_uint8,
      dhcp_opt_from_user_string_uint8,
      dhcp_opt_from_internal_string_uint8,
      dhcp_opt_destroy_atom,
    },

    /* Path MTU Aging Timeout Option */
    { TAG_DHCP_MTU_AGING_TIMEOUT,
      sizeof(uint32_t),
      DHCP_OPT_ATOM,
      DHCP_OPT_VAL_UINT32,
      dhcp_opt_to_network_uint32,
      dhcp_opt_to_user_string_uint32,
      dhcp_opt_to_internal_string_uint32,
      dhcp_opt_from_network_uint32,
      dhcp_opt_from_user_string_uint32,
      dhcp_opt_from_internal_string_uint32,
      dhcp_opt_destroy_atom,
    },

    /* Path MTU Plateau Table Option */
    { TAG_DHCP_MTU_PLATEAU_TABLE,
      sizeof(uint16_t),
      DHCP_OPT_LIST,
      DHCP_OPT_VAL_UINT16,
      dhcp_opt_to_network_uint16,
      dhcp_opt_to_user_string_uint16,
      dhcp_opt_to_internal_string_uint16,
      dhcp_opt_from_network_list_uint16,
      dhcp_opt_from_user_string_uint16,
      dhcp_opt_from_internal_string_uint16,
      dhcp_opt_destroy_list,
    },

    /* Interface MTU Option */
    { TAG_DHCP_INTERFACE_MTU,
      sizeof(uint16_t),
      DHCP_OPT_ATOM,
      DHCP_OPT_VAL_UINT16,
      dhcp_opt_to_network_uint16,
      dhcp_opt_to_user_string_uint16,
      dhcp_opt_to_internal_string_uint16,
      dhcp_opt_from_network_uint16,
      dhcp_opt_from_user_string_uint16,
      dhcp_opt_from_internal_string_uint16,
      dhcp_opt_destroy_atom,
    },

    /* Subnets Are Local Option */
    { TAG_DHCP_SUBNETS_ARE_LOCAL,
      sizeof(uint8_t),
      DHCP_OPT_ATOM,
      DHCP_OPT_VAL_UINT8,
      dhcp_opt_to_network_uint8,
      dhcp_opt_to_user_string_uint8,
      dhcp_opt_to_internal_string_uint8,
      dhcp_opt_from_network_uint8,
      dhcp_opt_from_user_string_uint8,
      dhcp_opt_from_internal_string_uint8,
      dhcp_opt_destroy_atom,
    },

    /* Broadcast Address Option */
    { TAG_DHCP_BROADCAST_ADDRESS,
      IP_ADDR_LEN,
      DHCP_OPT_ATOM,
      DHCP_OPT_VAL_ADDRESS,
      dhcp_opt_to_network_ip_addr,
      dhcp_opt_to_user_string_ip_addr,
      dhcp_opt_to_internal_string_ip_addr,
      dhcp_opt_from_network_ip_addr,
      dhcp_opt_from_user_string_ip_addr,
      dhcp_opt_from_internal_string_ip_addr,
      dhcp_opt_destroy_atom,
    },

    /* Perform Mask Discovery Option */
    { TAG_DHCP_MASK_DISCOVERY,
      sizeof(uint8_t),
      DHCP_OPT_ATOM,
      DHCP_OPT_VAL_UINT8,
      dhcp_opt_to_network_uint8,
      dhcp_opt_to_user_string_uint8,
      dhcp_opt_to_internal_string_uint8,
      dhcp_opt_from_network_uint8,
      dhcp_opt_from_user_string_uint8,
      dhcp_opt_from_internal_string_uint8,
      dhcp_opt_destroy_atom,
    },

    /* Mask Supplier */
    { TAG_DHCP_MASK_SUPPLIER,
      sizeof(uint8_t),
      DHCP_OPT_ATOM,
      DHCP_OPT_VAL_UINT8,
      dhcp_opt_to_network_uint8,
      dhcp_opt_to_user_string_uint8,
      dhcp_opt_to_internal_string_uint8,
      dhcp_opt_from_network_uint8,
      dhcp_opt_from_user_string_uint8,
      dhcp_opt_from_internal_string_uint8,
      dhcp_opt_destroy_atom,
    },

    /* Perform Router Discovery */
    { TAG_DHCP_ROUTER_DISCOVERY,
      sizeof(uint8_t),
      DHCP_OPT_ATOM,
      DHCP_OPT_VAL_UINT8,
      dhcp_opt_to_network_uint8,
      dhcp_opt_to_user_string_uint8,
      dhcp_opt_to_internal_string_uint8,
      dhcp_opt_from_network_uint8,
      dhcp_opt_from_user_string_uint8,
      dhcp_opt_from_internal_string_uint8,
      dhcp_opt_destroy_atom,
    },

    /* Router Solicitation Address Option */
    { TAG_DHCP_ROUTER_SOLICITATION,
      IP_ADDR_LEN,
      DHCP_OPT_ATOM,
      DHCP_OPT_VAL_ADDRESS,
      dhcp_opt_to_network_ip_addr,
      dhcp_opt_to_user_string_ip_addr,
      dhcp_opt_to_internal_string_ip_addr,
      dhcp_opt_from_network_ip_addr,
      dhcp_opt_from_user_string_ip_addr,
      dhcp_opt_from_internal_string_ip_addr,
      dhcp_opt_destroy_atom,
    },

    /* Static Route Option */
    { TAG_DHCP_STATIC_ROUTE,
      IP_ADDR_LEN,
      DHCP_OPT_LIST,
      DHCP_OPT_VAL_ADDRESS_PAIR,
      dhcp_opt_to_network_list_ip_addr_pair,
      dhcp_opt_to_user_string_list_ip_addr_pair,
      dhcp_opt_to_internal_string_list_ip_addr_pair,
      dhcp_opt_from_network_list_ip_addr_pair,
      dhcp_opt_from_user_string_ip_addr_pair,
      dhcp_opt_from_internal_string_list_ip_addr_pair,
      dhcp_opt_destroy_list,
    },

    /* Trailer Encapsulation Option */
    { TAG_DHCP_TRAILER_ENCAPSULATION,
      sizeof(uint8_t),
      DHCP_OPT_ATOM,
      DHCP_OPT_VAL_UINT8,
      dhcp_opt_to_network_uint8,
      dhcp_opt_to_user_string_uint8,
      dhcp_opt_to_internal_string_uint8,
      dhcp_opt_from_network_uint8,
      dhcp_opt_from_user_string_uint8,
      dhcp_opt_from_internal_string_uint8,
      dhcp_opt_destroy_atom,
    },

    /* Arp Cache Timeout */
    { TAG_DHCP_ARP_CACHE_TIMEOUT,
      sizeof(uint32_t),
      DHCP_OPT_ATOM,
      DHCP_OPT_VAL_UINT32,
      dhcp_opt_to_network_uint32,
      dhcp_opt_to_user_string_uint32,
      dhcp_opt_to_internal_string_uint32,
      dhcp_opt_from_network_uint32,
      dhcp_opt_from_user_string_uint32,
      dhcp_opt_from_internal_string_uint32,
      dhcp_opt_destroy_atom,
    },

    /* Ethernet Encapsulation */
    { TAG_DHCP_ETHERNET_ENCAPSULATION,
      sizeof(uint8_t),
      DHCP_OPT_ATOM,
      DHCP_OPT_VAL_UINT8,
      dhcp_opt_to_network_uint8,
      dhcp_opt_to_user_string_uint8,
      dhcp_opt_to_internal_string_uint8,
      dhcp_opt_from_network_uint8,
      dhcp_opt_from_user_string_uint8,
      dhcp_opt_from_internal_string_uint8,
      dhcp_opt_destroy_atom,
    },

    /* TCP Default TTL */
    { TAG_DHCP_TCP_DEFAULT_TTL,
      sizeof(uint8_t),
      DHCP_OPT_ATOM,
      DHCP_OPT_VAL_UINT8,
      dhcp_opt_to_network_uint8,
      dhcp_opt_to_user_string_uint8,
      dhcp_opt_to_internal_string_uint8,
      dhcp_opt_from_network_uint8,
      dhcp_opt_from_user_string_uint8,
      dhcp_opt_from_internal_string_uint8,
      dhcp_opt_destroy_atom,
    },

    /* TCP Keep Alive Interval */
    { TAG_DHCP_TCP_KEEPALIVE_INTERVAL,
      sizeof(uint32_t),
      DHCP_OPT_ATOM,
      DHCP_OPT_VAL_UINT32,
      dhcp_opt_to_network_uint32,
      dhcp_opt_to_user_string_uint32,
      dhcp_opt_to_internal_string_uint32,
      dhcp_opt_from_network_uint32,
      dhcp_opt_from_user_string_uint32,
      dhcp_opt_from_internal_string_uint32,
      dhcp_opt_destroy_atom,
    },

    /* Keep Alive Garbage */
    { TAG_DHCP_TCP_KEEPALIVE_GARBAGE,
      sizeof(uint8_t),
      DHCP_OPT_ATOM,
      DHCP_OPT_VAL_UINT8,
      dhcp_opt_to_network_uint8,
      dhcp_opt_to_user_string_uint8,
      dhcp_opt_to_internal_string_uint8,
      dhcp_opt_from_network_uint8,
      dhcp_opt_from_user_string_uint8,
      dhcp_opt_from_internal_string_uint8,
      dhcp_opt_destroy_atom,
    },

    /* NIS Domain */
    { TAG_DHCP_NIS_DOMAIN_NAME,
      sizeof(int8_t),
      DHCP_OPT_STRING,
      DHCP_OPT_VAL_INT8,
      dhcp_opt_to_network_nvt_string,
      dhcp_opt_to_user_string_nvt_string,
      dhcp_opt_to_internal_string_nvt_string,
      dhcp_opt_from_network_nvt_string,
      dhcp_opt_from_user_string_nvt_string,
      dhcp_opt_from_internal_string_nvt_string,
      dhcp_opt_destroy_string,
    },

    /* NIS Servers */
    { TAG_DHCP_NIS_SERVERS,
      IP_ADDR_LEN,
      DHCP_OPT_LIST,
      DHCP_OPT_VAL_ADDRESS,
      dhcp_opt_to_network_list_ip_addr,
      dhcp_opt_to_user_string_list_ip_addr,
      dhcp_opt_to_internal_string_list_ip_addr,
      dhcp_opt_from_network_list_ip_addr,
      dhcp_opt_from_user_string_list_ip_addr,
      dhcp_opt_from_internal_string_ip_addr,
      dhcp_opt_destroy_list,
    },

    /* NTP Servers */
    { TAG_DHCP_NTP_SERVERS,
      IP_ADDR_LEN,
      DHCP_OPT_LIST,
      DHCP_OPT_VAL_ADDRESS,
      dhcp_opt_to_network_list_ip_addr,
      dhcp_opt_to_user_string_list_ip_addr,
      dhcp_opt_to_internal_string_list_ip_addr,
      dhcp_opt_from_network_list_ip_addr,
      dhcp_opt_from_user_string_list_ip_addr,
      dhcp_opt_from_internal_string_list_ip_addr,
      dhcp_opt_destroy_list,
    },

    /* FIXME: this needs to be handled specially.  we actually
     * don't because for the time being we don't support any
     * vendor specific information. Good place to drop our own
     * home rolled options if we want. */

    /* Vendor Specific Information */
    { TAG_DHCP_VENDOR_SPECIFIC,
      sizeof(uint8_t),
      DHCP_OPT_ARRAY,
      DHCP_OPT_VAL_UINT8,
      dhcp_opt_to_network_array_uint8,
      dhcp_opt_to_user_string_array_uint8,
      dhcp_opt_to_internal_string_array_uint8,
      dhcp_opt_from_network_array_uint8,
      dhcp_opt_from_user_string_array_uint8,
      dhcp_opt_from_internal_string_array_uint8,
      dhcp_opt_destroy_array,
    },

    /* NetBIOS Name Server */
    { TAG_DHCP_NETBIOSOTCP_NAME_SERVER,
      IP_ADDR_LEN,
      DHCP_OPT_LIST,
      DHCP_OPT_VAL_ADDRESS,
      dhcp_opt_to_network_list_ip_addr,
      dhcp_opt_to_user_string_list_ip_addr,
      dhcp_opt_to_internal_string_list_ip_addr,
      dhcp_opt_from_network_list_ip_addr,
      dhcp_opt_from_user_string_list_ip_addr,
      dhcp_opt_from_internal_string_list_ip_addr,
      dhcp_opt_destroy_list,
    },

    /* NetBIOS Datagram Distribution Server */
    { TAG_DHCP_NETBIOSOTCP_DATAGRAM_SERVER,
      IP_ADDR_LEN,
      DHCP_OPT_LIST,
      DHCP_OPT_VAL_ADDRESS,
      dhcp_opt_to_network_list_ip_addr,
      dhcp_opt_to_user_string_list_ip_addr,
      dhcp_opt_to_internal_string_list_ip_addr,
      dhcp_opt_from_network_list_ip_addr,
      dhcp_opt_from_user_string_list_ip_addr,
      dhcp_opt_from_internal_string_list_ip_addr,
      dhcp_opt_destroy_list,
    },

    /* NetBIOS Datagram Node Type */
    { TAG_DHCP_NETBIOSOTCP_NODE_TYPE,
      sizeof(uint8_t),
      DHCP_OPT_ATOM,
      DHCP_OPT_VAL_UINT8,
      dhcp_opt_to_network_uint8,
      dhcp_opt_to_user_string_uint8,
      dhcp_opt_to_internal_string_uint8,
      dhcp_opt_from_network_uint8,
      dhcp_opt_from_user_string_uint8,
      dhcp_opt_from_internal_string_uint8,
      dhcp_opt_destroy_atom,
    },

    /* NetBIOS Scope Option */
    { TAG_DHCP_NETBIOSOTCP_SCOPE,
      sizeof(int8_t),
      DHCP_OPT_STRING,
      DHCP_OPT_VAL_INT8,
      dhcp_opt_to_network_string,
      dhcp_opt_to_user_string_string,
      dhcp_opt_to_internal_string_string,
      dhcp_opt_from_network_string,
      dhcp_opt_from_user_string_string,
      dhcp_opt_from_internal_string_string,
      dhcp_opt_destroy_atom,
    },

    /* X-Windows Font Server */
    { TAG_DHCP_XWIN_FONT_SERVER,
      IP_ADDR_LEN,
      DHCP_OPT_LIST,
      DHCP_OPT_VAL_ADDRESS,
      dhcp_opt_to_network_list_ip_addr,
      dhcp_opt_to_user_string_list_ip_addr,
      dhcp_opt_to_internal_string_list_ip_addr,
      dhcp_opt_from_network_list_ip_addr,
      dhcp_opt_from_user_string_list_ip_addr,
      dhcp_opt_from_internal_string_list_ip_addr,
      dhcp_opt_destroy_list,
    },

    /* X-Windows Display Manager Server */
    { TAG_DHCP_XWIN_DISP_MANAGER_SERVER,
      IP_ADDR_LEN,
      DHCP_OPT_LIST,
      DHCP_OPT_VAL_ADDRESS,
      dhcp_opt_to_network_list_ip_addr,
      dhcp_opt_to_user_string_list_ip_addr,
      dhcp_opt_to_internal_string_list_ip_addr,
      dhcp_opt_from_network_list_ip_addr,
      dhcp_opt_from_user_string_list_ip_addr,
      dhcp_opt_from_internal_string_list_ip_addr,
      dhcp_opt_destroy_list,
    },

    /* Requested IP Address */
    { TAG_DHCP_REQUESTED_IP_ADDRESS,
      IP_ADDR_LEN,
      DHCP_OPT_ATOM,
      DHCP_OPT_VAL_ADDRESS,
      dhcp_opt_to_network_ip_addr,
      dhcp_opt_to_user_string_ip_addr,
      dhcp_opt_to_internal_string_ip_addr,
      dhcp_opt_from_network_ip_addr,
      dhcp_opt_from_user_string_ip_addr,
      dhcp_opt_from_internal_string_ip_addr,
      dhcp_opt_destroy_atom,
    },

    /* IP Address Lease Time */
    { TAG_DHCP_IP_ADDRESS_LEASE_TIME,
      sizeof(uint32_t),
      DHCP_OPT_ATOM,
      DHCP_OPT_VAL_UINT32,
      dhcp_opt_to_network_uint32,
      dhcp_opt_to_user_string_uint32,
      dhcp_opt_to_internal_string_uint32,
      dhcp_opt_from_network_uint32,
      dhcp_opt_from_user_string_uint32,
      dhcp_opt_from_internal_string_uint32,
      dhcp_opt_destroy_atom,
    },

    /* Option Overload */
    { TAG_DHCP_OVERLOAD,
      sizeof(uint8_t),
      DHCP_OPT_ATOM,
      DHCP_OPT_VAL_UINT8,
      dhcp_opt_to_network_uint8,
      dhcp_opt_to_user_string_uint8,
      dhcp_opt_to_internal_string_uint8,
      dhcp_opt_from_network_uint8,
      dhcp_opt_from_user_string_uint8,
      dhcp_opt_from_internal_string_uint8,
      dhcp_opt_destroy_atom,
    },

    /* Message Type */
    { TAG_DHCP_MESSAGE_TYPE,
      sizeof(uint8_t),
      DHCP_OPT_ATOM,
      DHCP_OPT_VAL_UINT8,
      dhcp_opt_to_network_uint8,
      dhcp_opt_to_user_string_uint8,
      dhcp_opt_to_internal_string_uint8,
      dhcp_opt_from_network_uint8,
      dhcp_opt_from_user_string_uint8,
      dhcp_opt_from_internal_string_uint8,
      dhcp_opt_destroy_atom,
    },

    /* Server Identifier */
    { TAG_DHCP_SERVER_IDENTIFIER,
      IP_ADDR_LEN,
      DHCP_OPT_ATOM,
      DHCP_OPT_VAL_ADDRESS,
      dhcp_opt_to_network_ip_addr,
      dhcp_opt_to_user_string_ip_addr,
      dhcp_opt_to_internal_string_ip_addr,
      dhcp_opt_from_network_ip_addr,
      dhcp_opt_from_user_string_ip_addr,
      dhcp_opt_from_internal_string_ip_addr,
      dhcp_opt_destroy_atom,
    },

    /* FIXME: make a type for this. */
    /* Parameter Request List */
    { TAG_DHCP_PARAMETERS,
      sizeof(uint8_t),
      DHCP_OPT_ARRAY,
      DHCP_OPT_VAL_UINT8,
      dhcp_opt_to_network_array_uint8,
      dhcp_opt_to_user_string_array_uint8,
      dhcp_opt_to_internal_string_array_uint8,
      dhcp_opt_from_network_array_uint8,
      dhcp_opt_from_user_string_array_uint8,
      dhcp_opt_from_internal_string_array_uint8,
      dhcp_opt_destroy_array,
    },

    /* Message */
    { TAG_DHCP_MESSAGE,
      sizeof(int8_t),
      DHCP_OPT_STRING,
      DHCP_OPT_VAL_INT8,
      dhcp_opt_to_network_nvt_string,
      dhcp_opt_to_user_string_nvt_string,
      dhcp_opt_to_internal_string_nvt_string,
      dhcp_opt_from_network_nvt_string,
      dhcp_opt_from_user_string_nvt_string,
      dhcp_opt_from_internal_string_nvt_string,
      dhcp_opt_destroy_array,
    },

    /* Maximum DHCP Message Size */
    { TAG_DHCP_MAX_DHCP_SIZE,
      sizeof(uint16_t),
      DHCP_OPT_ATOM,
      DHCP_OPT_VAL_UINT16,
      dhcp_opt_to_network_uint16,
      dhcp_opt_to_user_string_uint16,
      dhcp_opt_to_internal_string_uint16,
      dhcp_opt_from_network_uint16,
      dhcp_opt_from_user_string_uint16,
      dhcp_opt_from_internal_string_uint16,
      dhcp_opt_destroy_atom,
    },

    /* Renewal Time */
    { TAG_DHCP_RENEWAL_TIME,
      sizeof(uint32_t),
      DHCP_OPT_ATOM,
      DHCP_OPT_VAL_UINT32,
      dhcp_opt_to_network_uint32,
      dhcp_opt_to_user_string_uint32,
      dhcp_opt_to_internal_string_uint32,
      dhcp_opt_from_network_uint32,
      dhcp_opt_from_user_string_uint32,
      dhcp_opt_from_internal_string_uint32,
      dhcp_opt_destroy_atom,
    },

    /* Rebinding Time */
    { TAG_DHCP_REBINDING_TIME,
      sizeof(uint32_t),
      DHCP_OPT_ATOM,
      DHCP_OPT_VAL_UINT32,
      dhcp_opt_to_network_uint32,
      dhcp_opt_to_user_string_uint32,
      dhcp_opt_to_internal_string_uint32,
      dhcp_opt_from_network_uint32,
      dhcp_opt_from_user_string_uint32,
      dhcp_opt_from_internal_string_uint32,
      dhcp_opt_destroy_atom,
    },

    /* Vendor class identifier */
    { TAG_DHCP_VENDOR_CLASS_ID,
      sizeof(uint8_t),
      DHCP_OPT_STRING,
      DHCP_OPT_VAL_UINT8,
      dhcp_opt_to_network_string,
      dhcp_opt_to_user_string_string,
      dhcp_opt_to_internal_string_string,
      dhcp_opt_from_network_string,
      dhcp_opt_from_user_string_string,
      dhcp_opt_from_internal_string_string,
      dhcp_opt_destroy_atom,
    },

    /* Client ID */
    { TAG_DHCP_CLIENT_ID,
      sizeof(uint8_t),
      DHCP_OPT_ARRAY,
      DHCP_OPT_VAL_UINT8,
      dhcp_opt_to_network_array_uint8,
      dhcp_opt_to_user_string_array_uint8,
      dhcp_opt_to_internal_string_array_uint8,
      dhcp_opt_from_network_array_uint8,
      dhcp_opt_from_user_string_array_uint8,
      dhcp_opt_from_internal_string_array_uint8,
      dhcp_opt_destroy_array,
    },

    /* NIS+ Domain */
    { TAG_DHCP_NISPLUS_DOMAIN,
      sizeof(int8_t),
      DHCP_OPT_STRING,
      DHCP_OPT_VAL_INT8,
      dhcp_opt_to_network_nvt_string,
      dhcp_opt_to_user_string_nvt_string,
      dhcp_opt_to_internal_string_nvt_string,
      dhcp_opt_from_network_nvt_string,
      dhcp_opt_from_user_string_nvt_string,
      dhcp_opt_from_internal_string_nvt_string,
      dhcp_opt_destroy_atom,
    },

    /* NIS+ Servers */
    { TAG_DHCP_NISPLUS_SERVERS,
      IP_ADDR_LEN,
      DHCP_OPT_LIST,
      DHCP_OPT_VAL_ADDRESS,
      dhcp_opt_to_network_list_ip_addr,
      dhcp_opt_to_user_string_list_ip_addr,
      dhcp_opt_to_internal_string_list_ip_addr,
      dhcp_opt_from_network_list_ip_addr,
      dhcp_opt_from_user_string_list_ip_addr,
      dhcp_opt_from_internal_string_list_ip_addr,
      dhcp_opt_destroy_list,
    },

    /* Mobile IP Home Agent */
    { TAG_DHCP_NISPLUS_SERVERS,
      IP_ADDR_LEN,
      DHCP_OPT_LIST,
      DHCP_OPT_VAL_ADDRESS,
      dhcp_opt_to_network_list_ip_addr,
      dhcp_opt_to_user_string_list_ip_addr,
      dhcp_opt_to_internal_string_list_ip_addr,
      dhcp_opt_from_network_list_ip_addr,
      dhcp_opt_from_user_string_list_ip_addr,
      dhcp_opt_from_internal_string_list_ip_addr,
      dhcp_opt_destroy_list,
    },

    /* TFTP Server Name Option */
    { TAG_DHCP_TFTP_SERVER,
      sizeof(int8_t),
      DHCP_OPT_STRING,
      DHCP_OPT_VAL_INT8,
      dhcp_opt_to_network_string,
      dhcp_opt_to_user_string_string,
      dhcp_opt_to_internal_string_string,
      dhcp_opt_from_network_string,
      dhcp_opt_from_user_string_string,
      dhcp_opt_from_internal_string_string,
      dhcp_opt_destroy_string,
    },

    /* Boot File Name Option */
    { TAG_DHCP_BOOT_FILE_NAME,
      sizeof(int8_t),
      DHCP_OPT_STRING,
      DHCP_OPT_VAL_INT8,
      dhcp_opt_to_network_string,
      dhcp_opt_to_user_string_string,
      dhcp_opt_to_internal_string_string,
      dhcp_opt_from_network_string,
      dhcp_opt_from_user_string_string,
      dhcp_opt_from_internal_string_string,
      dhcp_opt_destroy_string,
    },

    /* FIXME: this option can be legally set to 0
     * we need to handle it that way. */

    /* IP Home Agent Option */
    { TAG_DHCP_IP_HOME_AGENT,
      IP_ADDR_LEN,
      DHCP_OPT_LIST,
      DHCP_OPT_VAL_ADDRESS,
      dhcp_opt_to_network_list_ip_addr,
      dhcp_opt_to_user_string_list_ip_addr,
      dhcp_opt_to_internal_string_list_ip_addr,
      dhcp_opt_from_network_list_ip_addr,
      dhcp_opt_from_user_string_list_ip_addr,
      dhcp_opt_from_internal_string_list_ip_addr,
      dhcp_opt_destroy_list,
    },

    /* SMTP Server */
    { TAG_DHCP_SMTP_SERVER,
      IP_ADDR_LEN,
      DHCP_OPT_LIST,
      DHCP_OPT_VAL_ADDRESS,
      dhcp_opt_to_network_list_ip_addr,
      dhcp_opt_to_user_string_list_ip_addr,
      dhcp_opt_to_internal_string_list_ip_addr,
      dhcp_opt_from_network_list_ip_addr,
      dhcp_opt_from_user_string_list_ip_addr,
      dhcp_opt_from_internal_string_list_ip_addr,
      dhcp_opt_destroy_list,
    },

    /* POP Server */
    { TAG_DHCP_POP_SERVER,
      IP_ADDR_LEN,
      DHCP_OPT_LIST,
      DHCP_OPT_VAL_ADDRESS,
      dhcp_opt_to_network_list_ip_addr,
      dhcp_opt_to_user_string_list_ip_addr,
      dhcp_opt_to_internal_string_list_ip_addr,
      dhcp_opt_from_network_list_ip_addr,
      dhcp_opt_from_user_string_list_ip_addr,
      dhcp_opt_from_internal_string_list_ip_addr,
      dhcp_opt_destroy_list,
    },

    /* POP Server */
    { TAG_DHCP_POP_SERVER,
      IP_ADDR_LEN,
      DHCP_OPT_LIST,
      DHCP_OPT_VAL_ADDRESS,
      dhcp_opt_to_network_list_ip_addr,
      dhcp_opt_to_user_string_list_ip_addr,
      dhcp_opt_to_internal_string_list_ip_addr,
      dhcp_opt_from_network_list_ip_addr,
      dhcp_opt_from_user_string_list_ip_addr,
      dhcp_opt_from_internal_string_list_ip_addr,
      dhcp_opt_destroy_list,
    },

    /* NNTP Server */
    { TAG_DHCP_NNTP_SERVER,
      IP_ADDR_LEN,
      DHCP_OPT_LIST,
      DHCP_OPT_VAL_ADDRESS,
      dhcp_opt_to_network_list_ip_addr,
      dhcp_opt_to_user_string_list_ip_addr,
      dhcp_opt_to_internal_string_list_ip_addr,
      dhcp_opt_from_network_list_ip_addr,
      dhcp_opt_from_user_string_list_ip_addr,
      dhcp_opt_from_internal_string_list_ip_addr,
      dhcp_opt_destroy_list,
    },

    /* Default WWW Server */
    { TAG_DHCP_DEFAULT_WWW_SERVER,
      IP_ADDR_LEN,
      DHCP_OPT_LIST,
      DHCP_OPT_VAL_ADDRESS,
      dhcp_opt_to_network_list_ip_addr,
      dhcp_opt_to_user_string_list_ip_addr,
      dhcp_opt_to_internal_string_list_ip_addr,
      dhcp_opt_from_network_list_ip_addr,
      dhcp_opt_from_user_string_list_ip_addr,
      dhcp_opt_from_internal_string_list_ip_addr,
      dhcp_opt_destroy_list,
    },

    /* Default Finger Server */
    { TAG_DHCP_DEFAULT_FINGER_SERVER,
      IP_ADDR_LEN,
      DHCP_OPT_LIST,
      DHCP_OPT_VAL_ADDRESS,
      dhcp_opt_to_network_list_ip_addr,
      dhcp_opt_to_user_string_list_ip_addr,
      dhcp_opt_to_internal_string_list_ip_addr,
      dhcp_opt_from_network_list_ip_addr,
      dhcp_opt_from_user_string_list_ip_addr,
      dhcp_opt_from_internal_string_list_ip_addr,
      dhcp_opt_destroy_list,
    },

    /* Default IRC Server */
    { TAG_DHCP_DEFAULT_IRC_SERVER,
      IP_ADDR_LEN,
      DHCP_OPT_LIST,
      DHCP_OPT_VAL_ADDRESS,
      dhcp_opt_to_network_list_ip_addr,
      dhcp_opt_to_user_string_list_ip_addr,
      dhcp_opt_to_internal_string_list_ip_addr,
      dhcp_opt_from_network_list_ip_addr,
      dhcp_opt_from_user_string_list_ip_addr,
      dhcp_opt_from_internal_string_list_ip_addr,
      dhcp_opt_destroy_list,
    },

    /* Default Street Talk Server */
    { TAG_DHCP_DEFAULT_ST_SERVER,
      IP_ADDR_LEN,
      DHCP_OPT_LIST,
      DHCP_OPT_VAL_ADDRESS,
      dhcp_opt_to_network_list_ip_addr,
      dhcp_opt_to_user_string_list_ip_addr,
      dhcp_opt_to_internal_string_list_ip_addr,
      dhcp_opt_from_network_list_ip_addr,
      dhcp_opt_from_user_string_list_ip_addr,
      dhcp_opt_from_internal_string_list_ip_addr,
      dhcp_opt_destroy_list,
    },

    /* Default Street Talk Directory Server */
    { TAG_DHCP_DEFAULT_STDA_SERVER,
      IP_ADDR_LEN,
      DHCP_OPT_LIST,
      DHCP_OPT_VAL_ADDRESS,
      dhcp_opt_to_network_list_ip_addr,
      dhcp_opt_to_user_string_list_ip_addr,
      dhcp_opt_to_internal_string_list_ip_addr,
      dhcp_opt_from_network_list_ip_addr,
      dhcp_opt_from_user_string_list_ip_addr,
      dhcp_opt_from_internal_string_list_ip_addr,
      dhcp_opt_destroy_list,
    },

};

/* utilities. */

/* create a template: an option datum pointing to the appropriate attributes. */
static dhcp_opt_t *dhcp_opt_create_template(dhcp_opt_tag_t tag)
{
    dhcp_opt_t *opt;

    if(tag > MAX_OPTIONS_HANDLED)  {
        ERROR_MESSAGE("passed out of range option tag: %d", tag);
        return NULL;
    }

    opt = xcalloc(sizeof(dhcp_opt_t));
    opt->opt_attr = &dhcp_opt_attr[tag];

    return opt;
}

/* * * * * * * * * *
 * create/destroy. *
 * * * * * * * * * */

/* destruction routines. */

/* destroy an atom type value. */
static void dhcp_opt_destroy_atom(dhcp_opt_t *opt)
{
    if(opt->val != NULL)
        xfree(opt->val);

    return;
}

/* destroy a list type value. */
static void dhcp_opt_destroy_list(dhcp_opt_t *opt)
{
    list_t  *list;

    list = opt->val;

    if(list != NULL)
        list_destroy(list, xfree);

    return;
}

/* destroy a string type value. */
static void dhcp_opt_destroy_string(dhcp_opt_t *opt)
{
    return dhcp_opt_destroy_atom(opt);
}

/* destroy an array type value. */
static void dhcp_opt_destroy_array(dhcp_opt_t *opt)
{
    return dhcp_opt_destroy_atom(opt);
}

void dhcp_opt_destroy(dhcp_opt_t *opt)
{
    DHCP_OPT_DESTROY(opt);
    xfree(opt);

    return;
}

/* create routines. */

/* create a new dhcp option datum from network data. */
dhcp_opt_t *dhcp_opt_create_from_network(const uint8_t *data, dhcp_opt_len_t len, dhcp_opt_tag_t tag)
{
    dhcp_opt_t *opt;

    /* create a template. */
    opt = dhcp_opt_create_template(tag);

    if(opt == NULL)
        return NULL;

    /* now that we have a template populate the data. */
    if(DHCP_OPT_FROM_NETWORK_DATA(opt, data, len))  {
        dhcp_opt_destroy(opt);
        return NULL;
    }

    return opt;
}

/* create from internal data. */
dhcp_opt_t *dhcp_opt_create_from_internal_data(uint8_t tag, void *data, size_t len)
{
    dhcp_opt_t *opt;

    /* create a template. */
    opt = dhcp_opt_create_template(tag);

    if(opt == NULL)
        return NULL;

    switch(DHCP_OPT_GET_TYPE(opt))  {

    case DHCP_OPT_LIST:

        /* if its a list we need to do a list copy. otherwise just
         * copy the memory over. arg len doesn't apply here. */

        opt->val = list_copy(data, DHCP_OPT_GET_MEM_LEN(opt));
        opt->num = list_get_len(data);
        break;

    case DHCP_OPT_ATOM:

        /* an atom means a single datum the size of the option.
         * copy the memory over. arg len doesn't apply here. */

        opt->val = xmalloc(DHCP_OPT_GET_MEM_LEN(opt));
        memcpy(opt->val, data, DHCP_OPT_GET_MEM_LEN(opt));
        opt->num = 1;
        break;

    case DHCP_OPT_ARRAY:

        /* array means multiple datums which are better stored in
         * a contiginous space of memory (not a list).  len is
         * the total size, so we need to divide by the size of
         * the datum to find num. */

        opt->val = xmalloc(len);
        opt->num = len/DHCP_OPT_GET_MEM_LEN(opt);

        memcpy(opt->val, data, len);
        break;

    case DHCP_OPT_STRING:

        /* a string is easily copied. since it's internal we
         * expect it to be null terminated. */

        opt->val = xstrdup(data);
        opt->num = strlen(data);

        break;

    default:
        FATAL_MESSAGE("illegal type indicated by dhcp option. this is a bug report me.");
        exit(1);
    }

    return opt;
}

/* create from a user supplied string. */
dhcp_opt_t *dhcp_opt_create_from_user_string(uint8_t tag, void *string_data)
{
    dhcp_opt_t *opt;

    /* create a template. */
    opt = dhcp_opt_create_template(tag);

    if(opt == NULL)
        return NULL;

    if(DHCP_OPT_FROM_USER_STRING(opt, string_data))  {
        dhcp_opt_destroy(opt);
        return NULL;
    }

    return opt;
}

/* create from an internally formed string. */
dhcp_opt_t *dhcp_opt_create_from_host_string(uint8_t tag, char *input)
{
    dhcp_opt_t *opt;

    /* create a template. */
    opt = dhcp_opt_create_template(tag);

    if(opt == NULL)
        return NULL;

    if(DHCP_OPT_FROM_INTERNAL_STRING(opt, input))  {
        dhcp_opt_destroy(opt);
        return NULL;
    }

    return opt;
}

/* get routines. */

/* get network data. */
void *dhcp_opt_get_network_data(dhcp_opt_t *opt)
{
    return DHCP_OPT_TO_NETWORK_DATA(opt);
}

/* get host data. */
void *dhcp_opt_get_host_data(dhcp_opt_t *opt)
{
    return opt->val;
}

/* get user string. */
char *dhcp_opt_get_user_string(dhcp_opt_t *opt)
{
    return DHCP_OPT_TO_USER_STRING(opt);
}

/* get internal string. */
char *dhcp_opt_get_internal_string(dhcp_opt_t *opt)
{
    return DHCP_OPT_TO_INTERNAL_STRING(opt);
}

/* get member length. */
size_t dhcp_opt_get_mem_len(dhcp_opt_t *opt)
{
    return DHCP_OPT_GET_MEM_LEN(opt);
}

/* get total length. */
dhcp_opt_len_t dhcp_opt_get_total_len(dhcp_opt_t *opt)
{
    return DHCP_OPT_GET_TOTAL_LEN(opt);
}

/* get num. */
size_t dhcp_opt_get_num(dhcp_opt_t *opt)
{
    return opt->num;
}

dhcp_opt_type_t dhcp_opt_get_type(dhcp_opt_t *opt)
{
    return DHCP_OPT_GET_TYPE(opt);
}

dhcp_opt_type_t dhcp_opt_get_val_type(dhcp_opt_t *opt)
{
    return DHCP_OPT_GET_VAL_TYPE(opt);
}

dhcp_opt_tag_t dhcp_opt_get_tag(dhcp_opt_t *opt)
{
    return DHCP_OPT_GET_TAG(opt);
}

/* convenience routines. */

/* create a dhcp message type option. */
dhcp_opt_t *dhcp_opt_create_message_type(uint8_t message_type)
{
    return dhcp_opt_create_from_internal_data(TAG_DHCP_MESSAGE_TYPE, &message_type, sizeof(message_type));
}

/* accept bit array telling us which options to build the request option. */
dhcp_opt_t *dhcp_opt_create_parameter_request_list(uint8_t *requested_options)
{
    uint8_t i, j;
    uint8_t len = 0;
    dhcp_opt_t *option;
    uint8_t *opt_request;

    /* Grab length so we can just malloc all at once. */
    for(i = 0; i < MAX_OPTIONS_HANDLED; i++) {
        if(requested_options[i])
            len++;
    }

    if(len == 0) {
        /* no parameter list. return empty list. */
        return NULL;
    }

    opt_request = xmalloc(len * sizeof(uint8_t));

    j = 0;
    for(i = 0; i < MAX_OPTIONS_HANDLED; i++) {
        if(requested_options[i]) {
            opt_request[j] = i;
            j++;
        }
    }

    option = dhcp_opt_create_from_internal_data(TAG_DHCP_PARAMETERS, opt_request, len);
    xfree(opt_request);

    return option;
}

/* destroy a list of options by calling their destruction routine. */
void dhcp_opt_destroy_option_list(list_t *list)
{
    dhcp_opt_t *option;

    list_rewind(list);
    while((option = list_next(list)) != NULL)  {

        list_remove_by_datum(list, option);
        dhcp_opt_destroy(option);
    }

    list_destroy(list, NULL);

    return;
}

/* copy a dhcp option. */
dhcp_opt_t *dhcp_option_copy(dhcp_opt_t *option)
{
    return dhcp_opt_create_from_internal_data(dhcp_opt_get_tag(option), 
                                              dhcp_opt_get_host_data(option),
                                              dhcp_opt_get_total_len(option));
}

/* append to a dhcp list datum. */
void dhcp_option_append(dhcp_opt_t *orig, dhcp_opt_t *data)
{

    uint8_t *new_data, *orig_data;
    list_t *orig_list, *append_list;
    dhcp_opt_len_t len;

    orig_list = dhcp_opt_get_host_data(orig);
    append_list = dhcp_opt_get_host_data(data);
    len = dhcp_opt_get_mem_len(orig);

    list_rewind(append_list);
    while((orig_data = list_next(append_list)) != NULL) {

        new_data = xmalloc(len);
        memcpy(new_data, orig_data, len);

        list_add_to_end(orig_list, new_data);
        orig->num++;
    }

    return;
}

/* prepend to a dhcp list datum. */
void dhcp_option_prepend(dhcp_opt_t *orig, dhcp_opt_t *data)
{

    uint8_t *new_data, *orig_data;
    list_t *orig_list, *prepend_list;
    dhcp_opt_len_t len;

    orig_list = dhcp_opt_get_host_data(orig);
    prepend_list = dhcp_opt_get_host_data(data);
    len = dhcp_opt_get_mem_len(orig);

    list_rewind(prepend_list);
    while((orig_data = list_next(prepend_list)) != NULL) {

        new_data = xmalloc(len);
        memcpy(new_data, orig_data, len);

        list_add(orig_list, new_data);
        orig->num++;
    }

    return;
}
